import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib import patches
import seaborn as sns
from datetime import datetime
import matplotlib.cm as cm
import numpy as np
import math
import os

# qualitative 12 colors : http://colorbrewer2.org/?type=qualitative&scheme=Paired&n=12 + 11 diverting
BREWER_COLORS = ['#a6cee3', '#1f78b4', '#b2df8a', '#33a02c', '#fb9a99', '#e31a1c', '#fdbf6f',
                 '#ff7f00', '#cab2d6', '#6a3d9a', '#ffff99', '#b15928', '#a50026', '#d73027',
                 '#f46d43', '#fdae61', '#fee090', '#ffffbf', '#e0f3f8', '#abd9e9',
                 '#74add1', '#4575b4', '#313695']

def plot_hist_distribution(distribution_data, description, param=None, values_to_scatter=None,
                           xticks_labelsize=10, yticks_labelsize=10, x_label_font_size=15, y_label_font_size=15,
                           labels=None, scatter_shapes=None, colors=None, tight_x_range=False,
                           twice_more_bins=False, background_color="black", labels_color="white",
                           xlabel="", ylabel=None, path_results=None, save_formats="pdf",
                           v_line=None, x_range=None,
                           ax_to_use=None, color_to_use=None):
    """
    Plot a distribution in the form of an histogram, with option for adding some scatter values
    :param distribution_data:
    :param description:
    :param param:
    :param values_to_scatter:
    :param labels:
    :param scatter_shapes:
    :param colors:
    :param tight_x_range:
    :param twice_more_bins:
    :param xlabel:
    :param ylabel:
    :param save_formats:
    :return:
    """
    distribution = np.array(distribution_data)
    if color_to_use is None:
        hist_color = "blue"
    else:
        hist_color = color_to_use
    edge_color = "white"
    if x_range is not None:
        min_range = x_range[0]
        max_range = x_range[1]
    elif tight_x_range:
        max_range = np.max(distribution)
        min_range = np.min(distribution)
    else:
        max_range = 100
        min_range = 0
    weights = (np.ones_like(distribution) / (len(distribution))) * 100
    if ax_to_use is None:
        fig, ax1 = plt.subplots(nrows=1, ncols=1,
                                gridspec_kw={'height_ratios': [1]},
                                figsize=(12, 12))
        ax1.set_facecolor(background_color)
        fig.patch.set_facecolor(background_color)
    else:
        ax1 = ax_to_use
    bins = int(np.sqrt(len(distribution)))
    if twice_more_bins:
        bins *= 2
    hist_plt, edges_plt, patches_plt = ax1.hist(distribution, bins=bins, range=(min_range, max_range),
                                                facecolor=hist_color,
                                                edgecolor=edge_color,
                                                weights=weights, log=False, label=description)
    if values_to_scatter is not None:
        scatter_bins = np.ones(len(values_to_scatter), dtype="int16")
        scatter_bins *= -1

        for i, edge in enumerate(edges_plt):
            # print(f"i {i}, edge {edge}")
            if i >= len(hist_plt):
                # means that scatter left are on the edge of the last bin
                scatter_bins[scatter_bins == -1] = i - 1
                break

            if len(values_to_scatter[values_to_scatter <= edge]) > 0:
                if (i + 1) < len(edges_plt):
                    bool_list = values_to_scatter < edge  # edges_plt[i + 1]
                    for i_bool, bool_value in enumerate(bool_list):
                        if bool_value:
                            if scatter_bins[i_bool] == -1:
                                new_i = max(0, i - 1)
                                scatter_bins[i_bool] = new_i
                else:
                    bool_list = values_to_scatter < edge
                    for i_bool, bool_value in enumerate(bool_list):
                        if bool_value:
                            if scatter_bins[i_bool] == -1:
                                scatter_bins[i_bool] = i

        decay = np.linspace(1.1, 1.15, len(values_to_scatter))
        for i, value_to_scatter in enumerate(values_to_scatter):
            if i < len(labels):
                ax1.scatter(x=value_to_scatter, y=hist_plt[scatter_bins[i]] * decay[i], marker=scatter_shapes[i],
                            color=colors[i], s=60, zorder=20, label=labels[i])
            else:
                ax1.scatter(x=value_to_scatter, y=hist_plt[scatter_bins[i]] * decay[i], marker=scatter_shapes[i],
                            color=colors[i], s=60, zorder=20)
    y_min, y_max = ax1.get_ylim()
    if v_line is not None:
        ax1.vlines(v_line, y_min, y_max,
                   color="white", linewidth=2,
                   linestyles="dashed", zorder=5)

    ax1.legend()

    if tight_x_range:
        ax1.set_xlim(min_range, max_range)
    else:
        ax1.set_xlim(0, 100)
        xticks = np.arange(0, 110, 10)

        ax1.set_xticks(xticks)
        # sce clusters labels
        ax1.set_xticklabels(xticks)
    ax1.yaxis.set_tick_params(labelsize=xticks_labelsize)
    ax1.xaxis.set_tick_params(labelsize=yticks_labelsize)
    ax1.tick_params(axis='y', colors=labels_color)
    ax1.tick_params(axis='x', colors=labels_color)
    # TO remove the ticks but not the labels
    # ax1.xaxis.set_ticks_position('none')

    if ylabel is None:
        ax1.set_ylabel("Distribution (%)", fontsize=30, labelpad=20)
    else:
        ax1.set_ylabel(ylabel, fontsize=y_label_font_size, labelpad=20)
    ax1.set_xlabel(xlabel, fontsize=x_label_font_size, labelpad=20)

    ax1.xaxis.label.set_color(labels_color)
    ax1.yaxis.label.set_color(labels_color)

    # padding between ticks label and  label axis
    # ax1.tick_params(axis='both', which='major', pad=15)

    if ax_to_use is None:
        fig.tight_layout()
        if isinstance(save_formats, str):
            save_formats = [save_formats]
        if path_results is None:
            path_results = param.path_results
        time_str = ""
        if param is not None:
            time_str = param.time_str
        for save_format in save_formats:
            fig.savefig(f'{path_results}/{description}'
                        f'_{time_str}.{save_format}',
                        format=f"{save_format}",
                                facecolor=fig.get_facecolor())

        plt.close()

def plot_spikes_raster(spike_nums=None, title=None, file_name=None,
                       time_str=None,
                       spike_train_format=False,
                       y_ticks_labels=None,
                       y_ticks_labels_size=None,
                       y_ticks_labels_color="white",
                       x_ticks_labels_color="white",
                       x_ticks_labels=None,
                       x_ticks_labels_size=None,
                       x_ticks=None,
                       hide_x_labels=False,
                       figure_background_color="black",
                       without_ticks=True,
                       save_raster=False,
                       show_raster=False,
                       plot_with_amplitude=False,
                       activity_threshold=None,
                       save_formats="png",
                       span_area_coords=None,
                       span_area_colors=None,
                       span_area_only_on_raster=True,
                       alpha_span_area=0.5,
                       cells_to_highlight=None,
                       cells_to_highlight_colors=None,
                       color_peaks_activity=False,
                       horizontal_lines=None,
                       horizontal_lines_colors=None,
                       horizontal_lines_sytle=None,
                       horizontal_lines_linewidth=None,
                       vertical_lines=None,
                       vertical_lines_colors=None,
                       vertical_lines_sytle=None,
                       vertical_lines_linewidth=None,
                       scatters_on_traces=None,
                       scatters_on_traces_marker="*",
                       scatters_on_traces_size=5,
                       sliding_window_duration=1,
                       show_sum_spikes_as_percentage=False,
                       span_cells_to_highlight=None,
                       span_cells_to_highlight_colors=None,
                       spike_shape="|",
                       spike_shape_size=10,
                       raster_face_color='black',
                       cell_spikes_color='white',
                       activity_sum_plot_color="white",
                       activity_sum_face_color="black",
                       y_lim_sum_activity=None,
                       seq_times_to_color_dict=None,
                       link_seq_categories=None,
                       link_seq_color=None, min_len_links_seq=3,
                       link_seq_line_width=1, link_seq_alpha=1,
                       jitter_links_range=1,
                       display_link_features=True,
                       seq_colors=None, debug_mode=False,
                       axes_list=None,
                       SCE_times=None,
                       ylabel=None,
                       without_activity_sum=False,
                       spike_nums_for_activity_sum=None,
                       spikes_sum_to_use=None,
                       size_fig=None,
                       cmap_name="jet", traces=None,
                       display_traces=False,
                       display_spike_nums=True,
                       traces_lw=0.3,
                       path_results=None,
                       without_time_str_in_file_name=False,
                       desaturate_color_according_to_normalized_amplitude=False,
                       lines_to_display=None,
                       lines_color="white",
                       lines_width=1,
                       lines_band=0,
                       lines_band_color="white",
                       use_brewer_colors_for_traces=False,
                       dpi=100
                       ):
    """
    Plot or save a raster given a 2d array either binary representing onsets, peaks or rising time, or made of float
    to represents traces or encoding in onset/peaks/rising time a value.
    :param spike_nums: np.array of 2D, axis=1 (lines) represents the cells, the columns representing the spikes
    It could be binary, or containing the amplitude, if amplitudes values should be display put plot_with_amplitude
    to True
    :param spike_train_format: if True, means the data is a list of np.array, and then spike_nums[i][j] is
    a timestamps value as float
    :param title: title to be plot
    :param file_name: name of the file if save_raster is True
    :param save_raster: if True, the plot will be save. To do so param should not be None and contain a variable
    path_results that will indicated where to save file_name
    :param show_raster: if True, the plot will be shown
    :param plot_with_amplitude: to display a color bar representing the content values.
    :param activity_threshold: Int representing a threshold that will be display as a red line on the sum of activity
    subplot.
    :param save_formats: string or list of string representing the formats in which saving the raster.
    Exemple: "pdf" or ["pdf", "png"]
    :param span_area_coords: List of list of tuples of two float representing coords (x, x) of span band with a color
    corresponding to the one in span_area_colors
    :param span_area_colors: list of colors, same len as span_area_coords
    :param span_area_only_on_raster: if True, means the span won't be on the sum of activity on the sub-plot as well
    :param cells_to_highlight: cells index to span (y-axis) with special spikes color, list of int
    :param cells_to_highlight_colors: cells colors to span, same len as cells_to_span, list of string
    :param color_peaks_activity: if True, will span to the color of cells_to_highlight_colors each time at which a cell
    among cells_to_highlight will spike on the activity peak diagram
    :param horizontal_lines: list of float, representing the y coord at which trace horizontal lines
    :param horizontal_lines_colors: if horizontal_lines is not None, will set the colors of each line,
    list of string or color code
    :param horizontal_lines_style: give the style of the lines, string
    :param vertical_lines: list of float, representing the x coord at which trace vertical lines
    :param vertical__lines_colors: if horizontal_lines is not None, will set the colors of each line,
    list of string or color code
    :param vertical__lines_style: give the style of the lines, string
    :param vertical_lines_linewidth: linewidth of vertical_lines
    :param raster_face_color: the background color of the raster
    :param cell_spikes_color: the color of the spikes of the raster
    :param spike_shape: shape of the spike, "|", "*", "o"
    :param spike_shape_size: use for shape != of "|"
    :param seq_times_to_color_dict: None or a dict with as the key a tuple of int representing the cell index,
    and as a value a list of set, each set composed of int representing the times value at which the cell spike should
    be colored. It will be colored if there is indeed a spike at that time otherwise, the default color will be used.
    :param seq_colors: A dict, with key a tuple represening the indices of the seq and as value of colors,
    a color, should have the same keys as seq_times_to_color_dict
    :param link_seq_color: if not None, give the color with which link the spikes from a sequence. If not None,
    seq_colors will be ignored. could be a dict with key same tuple as seq_times_to_color_dict or a string and then
    we use the same color for all seq
    :param min_len_links_seq: minimum len of a seq for the links to be drawn
    :param axes_list if not None, give a list of axes that will be used, and be filled, but no figure will be created
    or saved then. Doesn't work yet is show_amplitude is True
    :param SCE_times:  a list of tuple corresponding to the first and last index of each SCE,
    (last index being included in the SCE). Will display the position of the SCE and their number above the activity
    diagram. If None, the overall time will be displayed. Need to be adapted to the format spike_numw or
    spike_train. Equivalent to span_are_coords
    :param without_activity_sum: if True, don't plot the sum of activity diagram, valid only if axes_list is not None
    :param spike_nums_for_activity_sum: if different that the one given for the raster, should be the
    same second dimension
    :param spikes_sum_to_use: an array of 1D, that will be use to display the sum of activity,
    :param size_fig: tuple of int
    :param cmap_name: "jet" by default, used if with_amplitude for the colormap
    :param traces if not None and display_traces is True, will display traces instead of a raster
    :param display_traces, if True display traces
    :param display_spike_nums, if False, won't display a raster using spike_nums
    :param traces_lw, default 0.3,  linewidth of the traces
    :param path_results: indicate where to save the plot, replace the param.path_results if it exists
    :param desaturate_color_according_to_normalized_amplitude: if True, spike_nums should be filled with float between
    0 and 1, representing the amplitude of the spike. And if a color is given for a cell, then it will be desaturate
    according to this value
    :param lines_to_display, dict that takes for a key a tuple of int representing 2 cells, and as value a list of tuple of 2 float
    representing the 2 extremities of a line between those 2 cells. By defualt, no lines
    :param lines_color="white": colors of lines_to_display
    :param lines_width=1: width of lines_to_display
    :param lines_band=0: if > 0, display a band around the line with transparency
    :param lines_band_color="white"
    :return:
    """

    # qualitative 12 colors : http://colorbrewer2.org/?type=qualitative&scheme=Paired&n=12
    # + 11 diverting
    brewer_colors = ['#a6cee3', '#1f78b4', '#b2df8a', '#33a02c', '#fb9a99', '#e31a1c', '#fdbf6f',
                     '#ff7f00', '#cab2d6', '#6a3d9a', '#ffff99', '#b15928', '#a50026', '#d73027',
                     '#f46d43', '#fdae61', '#fee090', '#ffffbf', '#e0f3f8', '#abd9e9',
                     '#74add1', '#4575b4', '#313695']
    brewer_colors = brewer_colors[::-1]

    if (spike_nums is None) and (traces is None):
        return

    if display_traces:
        if traces is None:
            return

    if spike_nums_for_activity_sum is None:
        spike_nums_for_activity_sum = spike_nums

    if plot_with_amplitude and spike_train_format:
        # not possible for now
        return

    if spike_nums is None:
        n_cells = len(traces)
    else:
        n_cells = len(spike_nums)

    if axes_list is None:
        if size_fig is None:
            size_fig = (15, 8)
        if not plot_with_amplitude:
            if without_activity_sum:
                fig, ax1 = plt.subplots(nrows=1, ncols=1, sharex=False,
                                        figsize=size_fig, dpi=dpi)
            else:
                sharex = True  # False if (SCE_times is None) else True
                fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, sharex=sharex,
                                               gridspec_kw={'height_ratios': [10, 2]},
                                               figsize=size_fig, dpi=dpi)
            fig.set_tight_layout({'rect': [0, 0, 1, 0.95], 'pad': 1.5, 'h_pad': 1.5})
        else:
            fig = plt.figure(figsize=size_fig, dpi=dpi)
            fig.set_tight_layout({'rect': [0, 0, 1, 1], 'pad': 1, 'h_pad': 1})
            outer = gridspec.GridSpec(1, 2, width_ratios=[100, 1])  # , wspace=0.2, hspace=0.2)
        fig.patch.set_facecolor(figure_background_color)
    else:
        if without_activity_sum:
            ax1 = axes_list[0]
        else:
            ax1, ax2 = axes_list

    if plot_with_amplitude:
        inner = gridspec.GridSpecFromSubplotSpec(2, 1,
                                                 subplot_spec=outer[0], height_ratios=[10, 2])
        # inner.tight_layout(fig, pad=0.1)
        ax1 = fig.add_subplot(inner[0])  # plt.Subplot(fig, inner[0])
        min_value = np.min(spike_nums)
        max_value = np.max(spike_nums)
        step_color_value = 0.1
        colors = np.r_[np.arange(min_value, max_value, step_color_value)]
        mymap = plt.get_cmap("jet")
        # get the colors from the color map
        my_colors = mymap(colors)

        # colors = plt.cm.hsv(y / float(max(y)))
        scalar_map = plt.cm.ScalarMappable(cmap=cmap_name, norm=plt.Normalize(vmin=min_value, vmax=max_value))
        # # fake up the array of the scalar mappable. Urgh…
        scalar_map._A = []

    # -------- end plot with amplitude ---------

    ax1.set_facecolor(raster_face_color)

    min_time = 0
    max_time = 0

    max_n_color = 10
    if display_traces:
        n_times = len(traces[0, :])
        zorder_traces = 21 + len(traces)
        for cell, trace in enumerate(traces):
            if use_brewer_colors_for_traces:
                color = brewer_colors[cell % len(brewer_colors)]
            else:
                color = cm.nipy_spectral(((cell % max_n_color) + 1) / (max_n_color + 1))
            ax1.plot(np.arange(n_times), trace + cell, lw=traces_lw, color=color, zorder=zorder_traces)
            if scatters_on_traces is not None:
                times_to_scatter = np.where(scatters_on_traces[cell, :])[0]
                ax1.scatter(times_to_scatter, trace[times_to_scatter] + cell,
                            color="white",
                            marker=scatters_on_traces_marker,
                            s=scatters_on_traces_size, zorder=zorder_traces - 1)
            zorder_traces -= 1
            line_beg_x = -1
            line_end_x = n_times + 1
            ax1.hlines(cell, line_beg_x, line_end_x, lw=0.1, linestyles="dashed", color=color, zorder=5)
    if cells_to_highlight is not None:
        cells_to_highlight = np.array(cells_to_highlight)
    if display_spike_nums:
        for cell, spikes in enumerate(spike_nums):
            if spike_train_format:
                if cell == 0:
                    min_time = np.min(spikes)
                else:
                    min_time = int(np.min((min_time, np.min(spikes))))
                max_time = int(np.ceil(np.max((max_time, np.max(spikes)))))
            # print(f"Neuron {y}, total spikes {len(np.where(neuron)[0])}, "
            #       f"nb > 2: {len(np.where(neuron>2)[0])}, nb < 2: {len(np.where(neuron[neuron<2])[0])}")
            if display_traces:
                # same color as traces
                color_neuron = cm.nipy_spectral(((cell % max_n_color) + 1) / (max_n_color + 1))
            else:
                color_neuron = cell_spikes_color
            if cells_to_highlight is not None:
                if cell in cells_to_highlight:
                    index = np.where(cells_to_highlight == cell)[0][0]
                    color_neuron = cells_to_highlight_colors[index]
            if spike_train_format:
                neuron_times = spikes
            else:
                neuron_times = np.where(spikes > 0)[0]
            if spike_shape != "|":
                if plot_with_amplitude:
                    ax1.scatter(neuron_times, np.repeat(cell, len(neuron_times)),
                                color=scalar_map.to_rgba(spikes[spikes > 0]),
                                marker=spike_shape,
                                s=spike_shape_size, zorder=20)
                elif desaturate_color_according_to_normalized_amplitude:
                    n_spikes = len(neuron_times)
                    colors_list = [sns.desaturate(x, p) for x, p in
                                   zip([color_neuron] * n_spikes, spikes[neuron_times])]
                    ax1.scatter(neuron_times, np.repeat(cell, len(neuron_times)),
                                color=colors_list,
                                marker=spike_shape,
                                s=spike_shape_size, zorder=20)
                else:
                    if display_traces:
                        y_values = traces[cell, neuron_times] + cell
                    else:
                        y_values = np.repeat(cell, len(neuron_times))
                    ax1.scatter(neuron_times, y_values, color=color_neuron, marker=spike_shape,
                                s=spike_shape_size, zorder=20)
            else:
                if desaturate_color_according_to_normalized_amplitude:
                    n_spikes = len(neuron_times)
                    colors_list = [sns.desaturate(x, p) for x, p in
                                   zip([color_neuron] * n_spikes, spikes[neuron_times])]
                    ax1.vlines(neuron_times, cell - .5, cell + .5, color=colors_list,
                               linewidth=spike_shape_size, zorder=20)
                elif plot_with_amplitude:
                    ax1.vlines(neuron_times, cell - .5, cell + .5, color=scalar_map.to_rgba(spikes[spikes > 0]),
                               linewidth=spike_shape_size, zorder=20)
                else:
                    ax1.vlines(neuron_times, cell - .5, cell + .5, color=color_neuron, linewidth=spike_shape_size,
                               zorder=20)
    if lines_to_display is not None:
        """
        lines_to_display=None,
                       lines_color="white",
                       lines_width=1,
                       lines_band=0,
                       lines_band_color="white"
                       dict that takes for a key a tuple of int representing 2 cells, and as value a list of tuple of 2 float
    representing the 2 extremities of a line between those 2 cells. By defualt, no lines
        """
        for cells_tuple, spike_times_list in lines_to_display.items():
            for spike_times in spike_times_list:
                ax1.plot(list(spike_times), list(cells_tuple),
                         color=lines_color,
                         linewidth=lines_width, zorder=30, alpha=1)
                if lines_band > 0:
                    xy = np.zeros((4, 2))
                    xy[0, 0] = spike_times[0] - lines_band
                    xy[0, 1] = cells_tuple[0]
                    xy[1, 0] = spike_times[1] - lines_band
                    xy[1, 1] = cells_tuple[1]
                    xy[2, 0] = spike_times[1] + lines_band
                    xy[2, 1] = cells_tuple[1]
                    xy[3, 0] = spike_times[0] + lines_band
                    xy[3, 1] = cells_tuple[0]
                    band_patch = patches.Polygon(xy=xy,
                                                 fill=True,
                                                 # linewidth=line_width,
                                                 facecolor=lines_band_color,
                                                 # edgecolor=edge_color,
                                                 alpha=0.4,
                                                 zorder=10)  # lw=2
                    ax1.add_patch(band_patch)
                    # ax1.fill_between(list(spike_times), np.array(cells_tuple)-lines_band,
                    #                  np.array(cells_tuple)+lines_band, facecolor=lines_band_color, alpha=0.4,
                    #                  zorder=10)
    if seq_times_to_color_dict is not None:
        seq_count = 0
        links_labels = []
        links_labels_color = []
        links_labels_y_coord = []
        if jitter_links_range > 0:
            nb_jitters = 10
            indices_rand_x = np.linspace(-jitter_links_range, jitter_links_range, nb_jitters)
            np.random.shuffle(indices_rand_x)
        for seq_indices, seq_times_list in seq_times_to_color_dict.items():
            nb_seq_times = 0
            for times_list_index, times_list in enumerate(seq_times_list):
                x_coord_to_link = []
                y_coord_to_link = []
                for time_index, t in enumerate(times_list):
                    cell_index = seq_indices[time_index]
                    # in case of this seq of cell would be used in a zoom version of the raster
                    if cell_index >= n_cells:
                        continue
                    # first we make sure the cell does spike at the given time
                    if spike_train_format:
                        if t not in spike_nums[cell_index]:
                            continue
                    else:
                        pass
                        if spike_nums[cell_index, t] == 0:
                            cell_for_msg = cell_index
                            if y_ticks_labels is not None:
                                cell_for_msg = y_ticks_labels[cell_index]
                            print(f"Not there: seq {times_list_index} cell {cell_for_msg} - {cell_index}, "
                                  f"time {t}")
                            continue
                        # print(f"## There: seq {times_list_index} cell {cell_index}, time {t}")
                    if link_seq_color is not None:
                        x_coord_to_link.append(t)
                        y_coord_to_link.append(cell_index)
                    else:
                        # if so, we draw the spike
                        if spike_shape != "|":
                            ax1.scatter(t, cell_index, color=seq_colors[seq_indices],
                                        marker=spike_shape,
                                        s=spike_shape_size, zorder=20)
                        else:
                            ax1.vlines(t, cell_index - .5, cell_index + .5, color=seq_colors[seq_indices],
                                       linewidth=1, zorder=20)
                if (link_seq_color is not None) and (len(x_coord_to_link) >= min_len_links_seq):
                    if isinstance(link_seq_color, str):
                        color_to_use = link_seq_color
                    elif isinstance(link_seq_color, dict):
                        color_to_use = link_seq_color[seq_indices]
                    else:
                        color_to_use = link_seq_color[seq_count % len(link_seq_color)]
                    x_coord_to_link = np.array(x_coord_to_link)
                    if jitter_links_range > 0:
                        jitter_to_add = indices_rand_x[seq_count % nb_jitters]
                    else:
                        jitter_to_add = 0
                    ax1.plot(x_coord_to_link + jitter_to_add, y_coord_to_link,
                             color=color_to_use,
                             linewidth=link_seq_line_width, zorder=30, alpha=link_seq_alpha)
                    nb_seq_times += 1
            if nb_seq_times > 0:
                category = ""
                if link_seq_categories is not None:
                    category = "*" * link_seq_categories[seq_indices]
                links_labels.append(f"l{len(seq_indices)}, r{nb_seq_times} {category}")
                links_labels_color.append(color_to_use)
                links_labels_y_coord.append((seq_indices[0] + seq_indices[-1]) / 2)
            seq_count += 1

    if display_traces:
        ax1.set_ylim(-0.5, n_cells + 4)
    else:
        ax1.set_ylim(-0.5, n_cells)
    if y_ticks_labels is not None:
        ax1.set_yticks(np.arange(n_cells))
        ax1.set_yticklabels(y_ticks_labels)
    if (x_ticks_labels is not None) and (x_ticks is not None):
        ax1.set_xticks(x_ticks)
        ax1.tick_params('x', length=2, width=0.5, which='both')
        ax1.set_xticklabels(x_ticks_labels, rotation=45)  # ha="right", va="center
    if x_ticks_labels_size is not None:
        ax1.xaxis.set_tick_params(labelsize=x_ticks_labels_size)
    if y_ticks_labels_size is not None:
        ax1.yaxis.set_tick_params(labelsize=y_ticks_labels_size)
    else:
        if n_cells < 50:
            y_ticks_labels_size = 5
        elif n_cells < 100:
            y_ticks_labels_size = 4
        elif n_cells < 200:
            y_ticks_labels_size = 3
        elif n_cells < 400:
            y_ticks_labels_size = 1
        else:
            y_ticks_labels_size = 0.1
        ax1.yaxis.set_tick_params(labelsize=y_ticks_labels_size)
    if without_ticks:
        if x_ticks is not None:
            ax1.tick_params(axis='y', which='both', length=0)
        else:
            ax1.tick_params(axis='both', which='both', length=0)

    if seq_times_to_color_dict is not None:
        if link_seq_color is not None:
            ax_right = ax1.twinx()
            ax_right.set_frame_on(False)
            ax_right.set_ylim(-1, n_cells)
            ax_right.set_yticks(links_labels_y_coord)
            # clusters labels
            ax_right.set_yticklabels(links_labels)
            ax_right.yaxis.set_ticks_position('none')
            if y_ticks_labels_size > 1:
                y_ticks_labels_size -= 1
            else:
                y_ticks_labels_size -= 0.5
            ax_right.yaxis.set_tick_params(labelsize=y_ticks_labels_size)
            # ax_right.yaxis.set_tick_params(labelsize=2)
            for index in np.arange(len(links_labels)):
                ax_right.get_yticklabels()[index].set_color(links_labels_color[index])

    if spike_train_format:
        n_times = int(math.ceil(max_time - min_time))
    else:
        if spike_nums is None:
            n_times = traces.shape[1]
        else:
            n_times = len(spike_nums[0, :])

    # draw span to highlight some periods
    if span_area_coords is not None:
        if len(span_area_coords) != len(span_area_colors):
            raise Exception("span_area_coords and span_area_colors are not the same size")
        for index, span_area_coord in enumerate(span_area_coords):
            for coord in span_area_coord:
                if span_area_colors is not None:
                    color = span_area_colors[index]
                else:
                    color = "lightgrey"
                ax1.axvspan(coord[0], coord[1], alpha=alpha_span_area, facecolor=color, zorder=1)

    if (span_cells_to_highlight is not None):
        for index, cell_to_span in enumerate(span_cells_to_highlight):
            ax1.axhspan(cell_to_span - 0.5, cell_to_span + 0.5, alpha=0.4,
                        facecolor=span_cells_to_highlight_colors[index])

    if horizontal_lines is not None:
        line_beg_x = 0
        line_end_x = 0
        if spike_train_format:
            line_beg_x = min_time - 1
            line_end_x = max_time + 1
        else:
            line_beg_x = -1
            line_end_x = n_times + 1
        if horizontal_lines_linewidth is None:
            ax1.hlines(horizontal_lines, line_beg_x, line_end_x, color=horizontal_lines_colors, linewidth=2,
                       linestyles=horizontal_lines_sytle)
        else:
            ax1.hlines(horizontal_lines, line_beg_x, line_end_x, color=horizontal_lines_colors,
                       linewidth=horizontal_lines_linewidth,
                       linestyles=horizontal_lines_sytle)

    if vertical_lines is not None:
        line_beg_y = 0
        line_end_y = n_cells - 1
        ax1.vlines(vertical_lines, line_beg_y, line_end_y, color=vertical_lines_colors,
                   linewidth=vertical_lines_linewidth,
                   linestyles=vertical_lines_sytle)

    if spike_train_format:
        ax1.set_xlim(min_time - 1, max_time + 1)
    else:
        ax1.set_xlim(-1, n_times + 1)
    # ax1.margins(x=0, tight=True)

    if not without_activity_sum or hide_x_labels:
        ax1.get_xaxis().set_visible(False)

    if title is not None:
        ax1.set_title(title)
    # Give x axis label for the spike raster plot
    # ax.xlabel('Frames')
    # Give y axis label for the spike raster plot
    if ylabel is not None:
        ax1.set_ylabel(ylabel)

    # ax1.spines['left'].set_color(y_ticks_labels_color)
    # ax1.spines['bottom'].set_color(x_ticks_labels_color)
    # ax1.yaxis.label.set_color(y_ticks_labels_color)
    if isinstance(y_ticks_labels_color, list):
        for xtick, color in zip(ax1.get_yticklabels(), y_ticks_labels_color):
            xtick.set_color(color)
    else:
        ax1.tick_params(axis='y', colors=y_ticks_labels_color)
    ax1.tick_params(axis='x', colors=x_ticks_labels_color)

    if (axes_list is not None) and without_activity_sum:
        return

    if not without_activity_sum:
        # ################################################################################################
        # ################################ Activity sum plot part ################################
        # ################################################################################################
        if (sliding_window_duration >= 1) and (spikes_sum_to_use is None):
            sum_spikes = np.zeros(n_times)
            if spike_train_format:
                windows_sum = np.zeros((n_cells, n_times), dtype="int16")
                # one cell can participate to max one spike by window
                # if value is True, it means this cell has already been counted
                cell_window_participation = np.zeros((n_cells, n_times), dtype="bool")
                for cell, spikes_train in enumerate(spike_nums_for_activity_sum):
                    for spike_time in spikes_train:
                        # first determining to which windows to add the spike
                        spike_index = int(spike_time - min_time)
                        first_index_window = np.max((0, spike_index - sliding_window_duration))
                        if np.sum(cell_window_participation[cell, first_index_window:spike_index]) == 0:
                            windows_sum[cell, first_index_window:spike_index] += 1
                            cell_window_participation[cell, first_index_window:spike_index] = True
                        else:
                            for t in np.arange(first_index_window, spike_index):
                                if cell_window_participation[cell, t] is False:
                                    windows_sum[cell, t] += 1
                                    cell_window_participation[cell, t] = True
                sum_spikes = np.sum(windows_sum, axis=0)
                if debug_mode:
                    print("sliding window over")
                # for index, t in enumerate(np.arange(int(min_time), int((np.ceil(max_time) - sliding_window_duration)))):
                #     # counting how many cell fire during that window
                #     if (index % 1000) == 0:
                #         print(f"index {index}")
                #     sum_value = 0
                #     t_min = t
                #     t_max = t + sliding_window_duration
                #     for spikes_train in spike_nums:
                #         # give the indexes
                #         # np.where(np.logical_and(spikes_train >= t, spikes_train < t_max))
                #         spikes = spikes_train[np.logical_and(spikes_train >= t, spikes_train < t_max)]
                #         nb_spikes = len(spikes)
                #         if nb_spikes > 0:
                #             sum_value += 1
                #     sum_spikes[index] = sum_value
                # sum_spikes[(n_times - sliding_window_duration):] = sum_value
            else:
                for t in np.arange(0, (n_times - sliding_window_duration)):
                    # One spike by cell max in the sum process
                    sum_value = np.sum(spike_nums_for_activity_sum[:, t:(t + sliding_window_duration)], axis=1)
                    sum_spikes[t] = len(np.where(sum_value)[0])
                sum_spikes[(n_times - sliding_window_duration):] = len(np.where(sum_value)[0])
        elif spikes_sum_to_use is None:
            if spike_train_format:
                pass
            else:
                binary_spikes = np.zeros((n_cells, n_times), dtype="int8")
                for spikes, spikes in enumerate(spike_nums_for_activity_sum):
                    binary_spikes[spikes, spikes > 0] = 1
                # if (param is not None) and (param.bin_size > 1):
                #     sum_spikes = np.mean(np.split(np.sum(binary_spikes, axis=0), n_times // param.bin_size), axis=1)
                #     sum_spikes = np.repeat(sum_spikes, param.bin_size)
                # else:
                sum_spikes = np.sum(binary_spikes, axis=0)
        else:
            sum_spikes = spikes_sum_to_use

        if spike_train_format:
            x_value = np.arange(min_time, max_time)
        else:
            x_value = np.arange(n_times)

        if plot_with_amplitude:
            ax2 = fig.add_subplot(inner[1], sharex=ax1)

        ax2.set_facecolor(activity_sum_face_color)

        # sp = UnivariateSpline(x_value, sum_spikes, s=240)
        # ax2.fill_between(x_value, 0, smooth_curve(sum_spikes), facecolor="black") # smooth_curve(sum_spikes)
        if show_sum_spikes_as_percentage:
            if debug_mode:
                print("using percentages")
            sum_spikes = sum_spikes / n_cells
            sum_spikes *= 100
            if activity_threshold is not None:
                activity_threshold = activity_threshold / n_cells
                activity_threshold *= 100

        ax2.fill_between(x_value, 0, sum_spikes, facecolor=activity_sum_plot_color, zorder=10)
        if activity_threshold is not None:
            line_beg_x = 0
            line_end_x = 0
            if spike_train_format:
                line_beg_x = min_time - 1
                line_end_x = max_time + 1
            else:
                line_beg_x = -1
                line_end_x = len(spike_nums_for_activity_sum[0, :]) + 1
            ax2.hlines(activity_threshold, line_beg_x, line_end_x, color="red", linewidth=1, linestyles="dashed")

        # draw span to highlight some periods
        if (span_area_coords is not None) and (not span_area_only_on_raster):
            for index, span_area_coord in enumerate(span_area_coords):
                for coord in span_area_coord:
                    if span_area_colors is not None:
                        color = span_area_colors[index]
                    else:
                        color = "lightgrey"
                    ax2.axvspan(coord[0], coord[1], alpha=0.5, facecolor=color, zorder=1)

        # early born
        if (cells_to_highlight is not None) and color_peaks_activity:
            for index, cell_to_span in enumerate(cells_to_highlight):
                ax2.vlines(np.where(spike_nums_for_activity_sum[cell_to_span, :])[0], 0, np.max(sum_spikes),
                           color=cells_to_highlight_colors[index],
                           linewidth=2, linestyles="dashed", alpha=0.2)

        # ax2.yaxis.set_visible(False)
        ax2.set_frame_on(False)
        ax2.get_xaxis().set_visible(True)
        if y_lim_sum_activity is not None:
            ax2.set_ylim(y_lim_sum_activity[0], y_lim_sum_activity[1])
        else:
            ax2.set_ylim(0, np.max(sum_spikes))
        if spike_train_format:
            ax2.set_xlim(min_time - 1, max_time + 1)
        else:
            if spike_nums_for_activity_sum is not None:
                ax2.set_xlim(-1, len(spike_nums_for_activity_sum[0, :]) + 1)
            else:
                ax2.set_xlim(-1, len(spikes_sum_to_use) + 1)

        if SCE_times is not None:
            ax_top = ax2.twiny()
            ax_top.set_frame_on(False)
            if spike_train_format:
                ax_top.set_xlim(min_time - 1, max_time + 1)
            else:
                if spike_nums_for_activity_sum is not None:
                    ax_top.set_xlim(-1, len(spike_nums_for_activity_sum[0, :]) + 1)
                else:
                    ax_top.set_xlim(-1, len(spikes_sum_to_use) + 1)
            xticks_pos = []
            for times_tuple in SCE_times:
                xticks_pos.append(times_tuple[0])
            ax_top.set_xticks(xticks_pos)
            ax_top.xaxis.set_ticks_position('none')
            ax_top.set_xticklabels(np.arange(len(SCE_times)))
            ax_top.tick_params(axis='x', colors=x_ticks_labels_color)
            plt.setp(ax_top.xaxis.get_majorticklabels(), rotation=90)
            if len(SCE_times) > 30:
                ax_top.xaxis.set_tick_params(labelsize=3)
            elif len(SCE_times) > 50:
                ax_top.xaxis.set_tick_params(labelsize=2)
            elif len(SCE_times) > 100:
                ax_top.xaxis.set_tick_params(labelsize=1)
            elif len(SCE_times) > 300:
                ax_top.xaxis.set_tick_params(labelsize=0.5)
            else:
                ax_top.xaxis.set_tick_params(labelsize=4)
        # print(f"max sum_spikes {np.max(sum_spikes)}, mean  {np.mean(sum_spikes)}, median {np.median(sum_spikes)}")

        if without_ticks:
            ax2.tick_params(axis='both', which='both', length=0)
        # ax2.yaxis.label.set_color(y_ticks_labels_color)
        if isinstance(y_ticks_labels_color, list):
            for xtick, color in zip(ax2.get_yticklabels(), y_ticks_labels_color):
                xtick.set_color(color)
        else:
            ax2.tick_params(axis='y', colors=y_ticks_labels_color)
        ax2.tick_params(axis='x', colors=x_ticks_labels_color)

        if (x_ticks_labels is not None) and (x_ticks is not None):
            ax2.set_xticks(x_ticks)
            ax2.tick_params('x', length=2, width=0.5, which='both')
            ax2.set_xticklabels(x_ticks_labels, rotation=45)  # ha="right", va="center
        if x_ticks_labels_size is not None:
            ax2.xaxis.set_tick_params(labelsize=x_ticks_labels_size)

    # color bar section
    if plot_with_amplitude:
        inner_2 = gridspec.GridSpecFromSubplotSpec(1, 1,
                                                   subplot_spec=outer[1])  # , wspace=0.1, hspace=0.1)
        ax3 = fig.add_subplot(inner_2[0])  # plt.Subplot(fig, inner_2[0])
        cb = fig.colorbar(scalar_map, cax=ax3)
        cb.ax.tick_params(axis='y', colors="white")
    if axes_list is None:
        if save_raster and (path_results is not None):
            # transforming a string in a list
            if isinstance(save_formats, str):
                save_formats = [save_formats]
            if time_str is None:
                time_str = datetime.now().strftime("%Y_%m_%d.%H-%M-%S")
            for save_format in save_formats:
                if without_time_str_in_file_name:
                    fig.savefig(os.path.join(f'{path_results}', f'{file_name}.{save_format}'),
                                format=f"{save_format}",
                                facecolor=fig.get_facecolor())
                else:
                    fig.savefig(os.path.join(f'{path_results}', f'{file_name}_{time_str}.{save_format}'),
                                format=f"{save_format}",
                                facecolor=fig.get_facecolor())
        # Display the spike raster plot
        if show_raster:
            plt.show()
        plt.close()



from deepcinac.cinac_predictor import *
from deepcinac.cinac_structures import *
import tensorflow as tf
import numpy as np

"""
We're going to guide you on how using deepCINAC to either infer the neuronal activity or 
predict the cell type from your calcium imaging data. 

Here is a link to our gitlab page for more information about our package: https://gitlab.com/cossartlab/deepcinac

So far, to run this code, you will need some calcium imaging data to work on (in tiff format) and 
some segmentation data (ROIs) indicating the contours or pixels 
that compose your cells (compatible format are Caiman, suite2p, Fiji or NWB outputs).

The filenames of the data in the code corresponds to the demo data available on our gitlab repository, 
using suite2p ROIs data: https://gitlab.com/cossartlab/deepcinac/tree/master/demos/data

You will also need a model file (.json extension) and a file containing the weights of network (.h5 extensions). 
You can download some there: https://gitlab.com/cossartlab/deepcinac/tree/master/model

"""

"""
You can either choose to use the classifier to infer neuronal activity or to predict the cell type. 
To predict the cell type set the variable cell_type_classifier_mode to False, else True.
"""
if __name__ == '__main__':
    cell_type_classifier_mode = False

    # ------------------------- #
    #     Setting file names
    # ------------------------- #

    # root path, just used to avoid copying the path everywhere
    root_path = '/../demo'

    # path to calcium imaging data
    data_path = os.path.join(root_path, "data")
    movie_file_name = os.path.join(data_path, "demo_deepcinac_1.tif")

    # string used to identify the recording from which you want to predict activity
    identifier = "demo_deepcinac"

    # Path to your model data. It's possible to have more than one model, and use
    # each for different cell of the same recording (for exemple, one
    # network could be specialized for interneurons and the other one for pyramidal
    # cells)
    if cell_type_classifier_mode:
        weights_file_name = os.path.join(root_path,
                                         "model/cinac_cell_type_weights_22_2.h5")
        json_file_name = os.path.join(root_path,
                                      "model/cinac_cell_type_model_22_2.json")
    else:
        weights_file_name = os.path.join(root_path, "model/cinac_weights_v2.h5")
        json_file_name = os.path.join(root_path, "model/cinac_model_v2.json")

    # useful for cell type classifier
    cell_type_yaml_file = os.path.join(data_path, "cell_type_yaml_files",
                                       "pyr_vs_ins_multi_class.yaml")

    # path of the directory where the results will be save
    # a directory will be created each time the prediction is run
    # the directory name will be the date and time at which the analysis has been run
    # the predictions will be in this directory.
    results_path = os.path.join(root_path, "results")

    # not mandatory, just to test if you GPU is accessible
    device_name = tf.test.gpu_device_name()
    if device_name != '/device:GPU:0':
        raise SystemError('GPU device not found')
    print('Found GPU at: {}'.format(device_name))

    evaluate_cell_type_classifier = True

    if evaluate_cell_type_classifier and cell_type_classifier_mode:
        cinac_dir_name = os.path.join(data_path, "cinac_cell_type_ground_truth", "for_testing")

        evaluate_cell_type_predictions(cinac_dir_name, cell_type_yaml_file, results_path,
                                       json_file_name, weights_file_name, save_cell_type_distribution=True)
    else:
        # ############
        # Creating an instance of CinacRecording
        # this class will be use to link the calcium imaging movie and the ROIs.
        # ############

        cinac_recording = CinacRecording(identifier=identifier)

        # Creating and adding to cinac_recording the calcium imaging movie data
        cinac_movie = CinacTiffMovie(tiff_file_name=movie_file_name)

        # if you have the movie already loaded in memory (for exemple in an nwb file,
        # if not using external link), then if you could do instead:

        # cinac_movie = CinacTiffMovie(tiff_movie=tiff_movie)
        # tiff_movie being a 3d numpy array (n_frames*dim_y*dim_x)

        cinac_recording.set_movie(cinac_movie)

        """
        Adding the information regarding the ROIs to the CinacRecording instance.
        There are four options, de-comment the one you need
        """

        # -----------------------------
        # options 1: suite2p data
        # -----------------------------

        """
        Segmenting your data will produce npy files used to build the ROIs
        the file iscell.npy indicated which roi represents a real cell, only those will be used.
        stat.npy will conain the ROIs coordinates. 
        """

        is_cell_suite2p_file_name = os.path.join(data_path, "suite2p", "demo_deepcinac_iscell_1.npy")
        stat_suite2p_file_name = os.path.join(data_path, "suite2p", "demo_deepcinac_stat_1.npy")
        cinac_recording.set_rois_from_suite_2p(is_cell_file_name=is_cell_suite2p_file_name,
                                               stat_file_name=stat_suite2p_file_name)


        # ------------------------------------------------------
        # options 2: contours coordinate (such as CaImAn, Fiji)
        # ------------------------------------------------------

        """
        Args:
            coord: numpy array of 2d, first dimension of length 2 (x and y) and 2nd dimension 
                   of length the number of
                   cells. Could also be a list of lists or tuples of 2 integers
            from_matlab: Indicate if the data has been computed by matlab, 
                         then 1 will be removed to the coordinates so that it starts at zero.
            """

        # coord = None
        # cinac_recording.set_rois_2d_array(coord=coord, from_matlab=False)


        # ------------------------------------------------
        # options 3: NWB (Neurodata Without Borders) data
        # ------------------------------------------------

        """
        Args:
             nwb_data: nwb object instance
                 name_module: Name of the module to find segmentation. 
                 Used this way: nwb_data.modules[name_module]
                 Ex: name_module = 'ophys'
             name_segmentation: Name of the segmentation in which find the plane segmentation.
                 Used this way:get_plane_segmentation(name_segmentation)
                 Ex: name_segmentation = 'segmentation_suite2p'
             name_seg_plane: Name of the segmentation plane in which to find the ROIs data
                 Used this way: mod[name_segmentation]get_plane_segmentation(name_seq_plane)
                 Ex: name_segmentation = 'my_plane_seg'
        
        'pixel_mask' data need to be available in the segmentation plane for it to work
        
        """
        # nwb_data = None
        # name_module = ""
        # name_segmentation = ""
        # name_seg_plane = ""
        # cinac_recording.set_rois_from_nwb(nwb_data=nwb_data, name_module=name_module,
        #                                   name_segmentation=name_segmentation, name_seg_plane=name_seg_plane)


        # -----------------------
        # options 4: Pixel masks
        # -----------------------

        """
        Args:
            pixel_masks: list of list of 2 integers representing 
                         for each cell all the pixels that belongs to the cell
        This method is actually called by set_rois_from_nwb() after extracting the 
        pixel_masks data from the nwb_file
        """

        # pixel_masks = None
        # cinac_recording.set_rois_using_pixel_mask(pixel_masks=pixel_masks)

        """
        Then we decide which network will be used for predicting the cells' activity.
        
        A dictionnary with key a tuple of 3 elements is used.
        
        The 3 elements are:
        
        (string) the model file name (.json extension)
        (string) the weights of the network file name (.h5 extension)
        (string) identifier for this configuration, will be used to name the output file
        The dictionnary will contain as value the cells to be predicted by the key configuration. 
        If the value is set to None, then all the cells will be predicted using this configuration.
        """

        model_files_dict = dict()
        # predicting 10 first cells with this model, weights and string identifying the network
        model_files_dict[(json_file_name, weights_file_name, "demo_cinac")] = np.arange(10)

        """
        We now create an instance of CinacPredictor and add the CinacRecording we have just created.
        
        It's possible to add more than one instance of CinacRecording, they will be predicted on the same run then.
        
        The argument removed_cells_mapping allows to remove cells from the segmentation. 
        This could be useful as the network take in consideration the adjacent cells to predict the activity, 
        thus if a cell was wrongly added to segmentation, this could lower the accuracy of the classifier.
        """

        cinac_predictor = CinacPredictor()

        """
        Args:
        
            removed_cells_mapping: integers array of length the original numbers of 
                cells (such as defined in CinacRecording)
                and as value either of positive int representing the new index of 
                the cell or -1 if the cell has been removed
        """

        cinac_predictor.add_recording(cinac_recording=cinac_recording,
                                      removed_cells_mapping=None,
                                      model_files_dict=model_files_dict)

        """
        Finally, we run the prediction.
        
        The output format could be either a matlab file(.mat) and/or numpy one (.npy).
        
        If matlab is chosen, the predictions will be available under the key "predictions".
        
        The predictions are a 2d float array (n_cells * n_frames) with value between 0 and 1, representing the prediction of our classifier for each frame. 1 means the cell is 100% sure active at that time, 0 is 100% sure not active.
        
        A cell is considered active during the rising time of the calcium transient.
        
        We use a threshold of 0.5 to binarize the predictions array and make it a raster.
        """

        # you could decomment this line to make sure the GPU is used
        #with tf.device('/device:GPU:0'):

        # predictions are saved in the results_path and return as a dict,
        overlap_value = 0 if cell_type_classifier_mode else 0.5
        predictions_dict = cinac_predictor.predict(results_path=results_path,
                                                   output_file_formats="npy",
                                                   overlap_value=overlap_value,
                                                   cell_type_classifier_mode=cell_type_classifier_mode)

        # Do output_file_formats=["npy", "mat"] to get both extensions

        """
        Code to convert the predictions as binary raster
        """
        if not cell_type_classifier_mode:
            # dictionary with key identifier of the recording and value a binary 2d array
            binary_predictions_dict = dict()

            for identifier, predictions in predictions_dict.items():
                binary_predictions = np.zeros((len(predictions), len(predictions[0])),
                                              dtype="int8")
                binary_predictions[predictions > 0.5] = 1
                binary_predictions_dict[identifier] = binary_predictions



"""
We're going to guide you on how using DeepCINAC to generate a simulated version of 2-photon calcium imaging
(the goal is not to be super realistic).

Here is a link to our gitlab page (https://gitlab.com/cossartlab/deepcinac) for more information about our package.


So far, to run this code, you will need some cell contours model
(you can download a set from our gitlab at this address(https://gitlab.com/cossartlab/deepcinac/tree/master/demos/data/simulated_movies/cell_models_simulated_movie_suite2p.mat)
or follow the instructions below to generate your own.
You can also add vessels to your movie (see instructions below, you can dowload some here
(https://gitlab.com/cossartlab/deepcinac/tree/master/demos/data//vessel_pics) as an example ).
We haven't use vessels on our simulated data to train our classifier

You can check examples of simulated movie at this address:
(https://gitlab.com/cossartlab/deepcinac/tree/master/demos/data/simulated_movies).
"""

from deepcinac.cinac_simulated_movie_generator import SimulatedMovieGenerator
import os
from datetime import datetime

# root path, just used to avoid copying the path everywhere
root_path = ''

# path to calcium imaging data
data_path = os.path.join(root_path, "data")

# file containing cell models, based on suite2p segmentation from a few
# 2-photon calcium imaging recorded in the CA1 pyramidal layer on pups
cell_models_file = os.path.join(data_path,
                                "simulated_movie_generator",
                                "cell_models_simulated_movie_suite2p.mat")

# path of the directory where the results will be save
# a directory will be created each time the prediction is run
# the directory name will be the date and time at which the analysis has been run
# the predictions will be in this directory.
path_results = os.path.join(root_path, "results")
time_str = datetime.now().strftime("%Y_%m_%d.%H-%M-%S")
path_results = os.path.join(path_results, time_str)
if not os.path.isdir(path_results):
    os.mkdir(path_results)

# #### First step: create an instance of SimulatedMovieGenerator and decide of the parameters
# #### (dimensions, number of frames, transients' rate etc...)

"""
dimensions: tuple of 2 int, represent the dimension of the movie in pixels, (x, y)
n_overlap_by_cell_range: (tuple of 2 int) bottom and upper range of how many cells will overlap
            our cell "of interest" aimed to train the classifier
non_overlap_by_cell_range:  (tuple of 2 int) bottom and upper range of  how many cells will be near
            our cells "of interest" aimed to train the classifier, but without any overlap
n_cells_of_interest: (int) cells of interest, are the cell for which we want to control how many overlaps,
             their activity the aim is to use them to train the classifier
range_n_transient_cells_of_interest: bottom and upper range of the number of transients over 1000 frames
            for our cells of interest
range_n_transient_overlapping_cells: bottom and upper range of the number of transients over 1000 frames
            for cells that overlap our cell of interest
range_n_transient_other_cells: bottom and upper range of the number of transients over 1000 frames
            for cells that are not of interest or overlap them
with_mvt: boolean, is True, means some motion will be added, otherwise it would produce a movie with 
            perfect motion correction  
path_results: Directory in which to save the results
time_str (string): id (meant to be a timestamps string representation) in order to  give a unique id to files. 
"""
movie_generator = SimulatedMovieGenerator(dimensions=(120, 120),
                                          n_cells_of_interest=16,
                                          n_overlap_by_cell_range=(1, 4),
                                          non_overlap_by_cell_range=(2, 10),
                                          range_n_transient_cells_of_interest=(2, 4),
                                          range_n_transient_overlapping_cells=(8, 16),
                                          range_n_transient_other_cells=(2, 16),
                                          n_frames=2500,
                                          with_mvt=False,
                                          time_str=time_str,
                                          path_results=path_results)

# ## OPTIONAL CODE FOR INCLUDING VESSELS ###
"""
(optional) This code allows you to create vessels based on tiff images. You can find examples on our gitlab, 
it should be a picture with a white background and the vessel should be in black. 
"""
load_vessels=False

if load_vessels:

    vessels_imgs_dir = os.path.join(data_path, "simulated_movie_generator/vessels_tiff_imgs")

    # you can indicate a path where to save the newly created vessels using path_results to load
    # them directly another time
    # n_vessels_max is the number of vessels to add, if the directory contains more
    # than n_vessels_max

    movie_generator.produce_and_load_vessels(vessels_imgs_dir=vessels_imgs_dir,
                                             n_vessels_max=2, path_results=None)
    # ## Code to load vessels already created

    # vessels_dir is the directory that contains the vessels
    # n_vessels_max is the number of vessels to add, if the directory contains more
    # than n_vessels_max
    vessels_dir = os.path.join(data_path, "simulated_movie_generator")
    movie_generator.load_vessels(vessels_dir=vessels_dir, n_vessels_max=2)

# ## END VESSELS ###


# ## Then it is necessary to load some cell models that will be use to generate the contour map.
# from_matlab: boolean that indicates if the coordinates have been generated (indexing starts at 1 then)
# using matlab code and so starting from 1 and not 0.
movie_generator.load_cell_coords(data_file=cell_models_file, from_matlab=False)

"""
Finally, we generate the movie.

The automatic steps are:

1. Generating a contour map (a figure will be created to represent it as well as a file with the coordinates)
2. Generating a for each transients and building a raster 
(figure will be created as well as files to be opened in the GUI to check the activity)
3. Generating the traces (figure created as well)
4. Generating the movie (tiff file created)

"""

movie_generator.generate_movie()



import numpy as np
from deepcinac.utils.utils import get_continous_time_periods, find_all_onsets_and_peaks_on_fluorescence_signal

def neuronal_activity_encoding(raw_traces, smooth_traces, raster_dur, identifier=None):
    """
    Give for each frame of the cell what kind of activity is going on (real transient, fake etc...)
    Args:
        raw_traces: 1d float representing the raw fluorescence signal (should be normalized using z-score)
        smooth_traces: 1d float representing the smoothed fluorescence signal (should be normalized using z-score)
        raster_dur: 1d int (or bool), representing the frames during which a given cell is active
        (corresponding to the traces). It represents the "ground truth" used to know real transients.

    Returns:

    """
    n_frames = len(raw_traces)
    encoded_frames = np.zeros(n_frames, dtype="int16")
    decoding_frame_dict = dict()
    # zero will be the Neuropil
    decoding_frame_dict[0] = NeuropilEvent(frame_index=None)
    next_code = 1

    # first we add the real transient
    transient_periods = get_continous_time_periods(raster_dur)
    # print(f"sum ms.spike_struct.spike_nums_dur[cell] {np.sum(ms.spike_struct.spike_nums_dur[cell])}")
    # list of tuple, first frame and last frame (included) of each transient
    for transient_period in transient_periods:
        amplitude = np.max(raw_traces[transient_period[0]:transient_period[1] + 1])
        encoded_frames[transient_period[0]:transient_period[1] + 1] = next_code
        event = RealTransientEvent(frames_period=transient_period, amplitude=amplitude)
        decoding_frame_dict[next_code] = event
        next_code += 1

    # then we look for all transient and take the mean + 1 std of transient peak
    # and keep the one that are not real as fake one
    # for that first we need to compute peaks_nums, spike_nums and spike_nums_dur from all onsets
    raster_dur_all = find_all_onsets_and_peaks_on_fluorescence_signal(
        smooth_trace=smooth_traces, threshold_factor=1, identifier=identifier)
    all_transient_periods = get_continous_time_periods(raster_dur_all)
    for transient_period in all_transient_periods:
        # checking if it's part of a real transient
        sp_dur = raster_dur
        if np.sum(sp_dur[transient_period[0]:transient_period[1] + 1]) > 0:
            continue
        amplitude = np.max(raw_traces[transient_period[0]:transient_period[1] + 1])
        encoded_frames[transient_period[0]:transient_period[1] + 1] = next_code
        event = FakeTransientEvent(frames_period=transient_period, amplitude=amplitude)
        decoding_frame_dict[next_code] = event
        next_code += 1

    return encoded_frames, decoding_frame_dict


class MovieEvent:
    """
    Class that represent an event in a movie, for exemple a transient, neuropil etc...
    """

    def __init__(self):
        self.neuropil = False
        self.real_transient = False
        self.fake_transient = False
        self.movement = False
        # length in frames
        self.length_event = 1
        self.first_frame_event = None
        self.last_frame_event = None


class NeuropilEvent(MovieEvent):
    # it includes neuropil, but also decay of transient, everything that is not real or fake transient
    def __init__(self, frame_index):
        super().__init__()
        self.neuropil = True
        # frame_index could be None, in case we don't care about frame index
        self.first_frame_event = frame_index
        self.last_frame_event = frame_index


class RealTransientEvent(MovieEvent):

    def __init__(self, frames_period, amplitude):
        super().__init__()
        self.real_transient = True
        self.first_frame_event = frames_period[0]
        self.last_frame_event = frames_period[1]
        self.length_event = self.last_frame_event - self.first_frame_event + 1
        self.amplitude = amplitude


class FakeTransientEvent(MovieEvent):

    def __init__(self, frames_period, amplitude):
        super().__init__()
        self.fake_transient = True
        self.first_frame_event = frames_period[0]
        self.last_frame_event = frames_period[1]
        self.length_event = self.last_frame_event - self.first_frame_event + 1
        self.amplitude = amplitude


class MovementEvent(MovieEvent):

    def __init__(self, frames_period):
        super().__init__()
        self.movement = True
        self.first_frame_event = frames_period[0]
        self.last_frame_event = frames_period[1]
        self.length_event = self.last_frame_event - self.first_frame_event + 1


class StratificationCellTypeCamembert:

    def __init__(self, data_list, description,
                 n_max_transformations, debug_mode=False):
        self.data_list = data_list
        self.n_movie_patch = len(data_list)
        self.description = description
        self.debug_mode = debug_mode
        self.n_max_transformations = n_max_transformations
        if debug_mode:
            self.print_data_description()

    def augment_them_all(self):
        """
        Add to all movie patches in the camembert a given number of augmentation self.n_max_transformations
        :return:
        """
        for movie_patch_data in self.data_list:
            movie_patch_data.add_n_augmentation(self.n_max_transformations)

    def print_data_description(self):
        n_movie_patch = 0
        for movie_data in self.data_list:
            n_movie_patch += (1 + movie_data.n_augmentations_to_perform)

        # key is a string representing the cell type, and value an int representing the number of movie_patch of
        # this cell type
        cell_type_count_dict = dict()
        for movie_index, movie_data in enumerate(self.data_list):
            cell_type = movie_data.cinac_recording.cell_type
            cell_type_count_dict[cell_type] = cell_type_count_dict.get(cell_type, 0) + \
                                              (movie_data.n_augmentations_to_perform + 1)

        print("")
        print("-"*100)
        print("StratificationCellTypeCamembert")
        print(f"N movie patches: {n_movie_patch}")
        for cell_type, movie_patches_count in cell_type_count_dict.items():
            print(f"N movie patches for {cell_type}: {movie_patches_count}")
        print("-" * 100)
        print("")

class StratificationCamembert:

    def __init__(self, data_list, description,
                 n_max_transformations, debug_mode=False):
        self.data_list = data_list
        self.n_movie_patch = len(data_list)
        self.description = description
        self.debug_mode = debug_mode
        self.n_max_transformations = n_max_transformations

        ##### for fake and real transients
        self.n_transient_total = dict()
        self.n_transient_total["fake"] = 0
        self.n_transient_total["real"] = 0

        self.n_cropped_transient_total = dict()
        self.n_cropped_transient_total["fake"] = 0
        self.n_cropped_transient_total["real"] = 0

        self.n_full_transient_total = dict()
        self.n_full_transient_total["fake"] = 0
        self.n_full_transient_total["real"] = 0

        self.transient_movies = dict()
        self.transient_movies["fake"] = []
        self.transient_movies["real"] = []

        self.cropped_transient_movies = dict()
        self.cropped_transient_movies["fake"] = []
        self.cropped_transient_movies["real"] = []

        self.min_augmentation_for_transient = dict()
        self.min_augmentation_for_transient["fake"] = 2
        self.min_augmentation_for_transient["real"] = 2

        self.min_augmentation_for_cropped_transient = dict()
        self.min_augmentation_for_cropped_transient["fake"] = 0
        self.min_augmentation_for_cropped_transient["real"] = 0

        # count
        self.n_full_1_transient = dict()
        self.n_full_1_transient["fake"] = 0
        self.n_full_1_transient["real"] = 0

        self.n_full_2p_transient = dict()
        self.n_full_2p_transient["fake"] = 0
        self.n_full_2p_transient["real"] = 0

        # list of movie_data with full transient (1 rep)
        self.full_1_transient = dict()
        self.full_1_transient["fake"] = []
        self.full_1_transient["real"] = []

        self.full_2p_transient = dict()
        self.full_2p_transient["fake"] = []
        self.full_2p_transient["real"] = []

        # contains the indices of the movies (from data_list) in a sorted order (based on amplitude, from low to high)
        self.full_transient_sorted_amplitude = dict()
        self.full_transient_sorted_amplitude["fake"] = []
        self.full_transient_sorted_amplitude["real"] = []

        self.n_transient_dict = dict()
        self.n_transient_dict["fake"] = dict()
        self.n_transient_dict["real"] = dict()

        # perc of full with 1 transient
        self.full_1_transient_perc = dict()
        self.full_1_transient_perc["fake"] = 0
        self.full_1_transient_perc["real"] = 0

        # perc of full with 2 transients or more
        self.full_2p_transient_perc = dict()
        self.full_2p_transient_perc["fake"] = 0
        self.full_2p_transient_perc["real"] = 0

        # % among all transients
        self.full_transient_perc = dict()
        self.full_transient_perc["fake"] = 0
        self.full_transient_perc["real"] = 0

        self.cropped_transient_perc = dict()
        self.cropped_transient_perc["fake"] = 0
        self.cropped_transient_perc["real"] = 0

        self.n_cropped_transient_dict = dict()
        self.n_cropped_transient_dict["fake"] = dict()
        self.n_cropped_transient_dict["real"] = dict()

        self.total_transient_perc = dict()
        self.total_transient_perc["fake"] = 0
        self.total_transient_perc["real"] = 0

        self.transient_lengths = dict()
        self.transient_lengths["fake"] = []
        self.transient_lengths["real"] = []

        self.transient_amplitudes = dict()
        self.transient_amplitudes["fake"] = []
        self.transient_amplitudes["real"] = []

        # MoviePatchData list
        self.neuropil_movies = []

        self.n_only_neuropil = 0
        self.only_neuropil_perc = 0
        self.n_real_and_fake_transient = 0
        # disct with key ms.description and value the number of movies for this session
        self.n_movies_by_session = {}
        # key int representing age, and value the number of movie for this session
        self.n_movies_by_age = {}

        self.compute_slices()

    def compute_slices(self):
        """
        Compute the slices of the camembert
        :return:
        """
        # printing option
        # \x1b
        reset_color = '\033[30m'
        # red: 31, blue: 34
        perc_color = '\033[34m'
        self.n_movie_patch = 0
        for movie_data in self.data_list:
            self.n_movie_patch += (1 + movie_data.n_augmentations_to_perform)

        sorted_amplitudes = dict()
        amplitudes_movie_index = dict()

        for which_ones in ["real", "fake"]:
            # initializing variables
            self.transient_movies[which_ones] = []
            self.cropped_transient_movies[which_ones] = []
            self.full_1_transient[which_ones] = []
            self.full_2p_transient[which_ones] = []
            self.n_transient_dict[which_ones] = dict()
            self.n_cropped_transient_dict[which_ones] = dict()
            self.transient_lengths[which_ones] = []
            self.transient_amplitudes[which_ones] = []
            self.full_transient_sorted_amplitude[which_ones] = []
            sorted_amplitudes[which_ones] = []
            amplitudes_movie_index[which_ones] = []

        # MoviePatchData list
        self.neuropil_movies = []
        self.n_only_neuropil = 0
        self.only_neuropil_perc = 0
        self.n_real_and_fake_transient = 0
        # disct with key ms.description and value the number of movies for this session
        self.n_movies_by_session = {}
        # key int representing age, and value the number of movie for this session
        self.n_movies_by_age = {}

        if self.debug_mode:
            print(f"{'##' * 10}")
            print(f"{self.description}")
            print(f"{'##' * 10}")
        for movie_index, movie_data in enumerate(self.data_list):
            movie_info = movie_data.movie_info
            only_neuropil = True
            with_real_transient = False
            with_cropped_real_transient = False
            with_fake_transient = False
            n_movies = 1 + movie_data.n_augmentations_to_perform
            if "n_transient" in movie_info:
                with_real_transient = True
                only_neuropil = False
                n_transient = movie_info["n_transient"]
                if n_transient == 1:
                    self.full_1_transient["real"].append(movie_data)
                else:
                    self.full_2p_transient["real"].append(movie_data)
                self.transient_movies["real"].append(movie_data)
                self.n_transient_dict["real"][n_transient] = self.n_transient_dict["real"].get(n_transient,
                                                                                               0) + n_movies
                if "transients_amplitudes" in movie_info:
                    self.transient_amplitudes["real"].extend(movie_info["transients_amplitudes"])
                    sorted_amplitudes["real"].append(np.max(movie_info["transients_amplitudes"]))
                    amplitudes_movie_index["real"].append(movie_index)
                if "transients_lengths" in movie_info:
                    self.transient_lengths["real"].extend(movie_info["transients_lengths"])
            if ("n_cropped_transient" in movie_info) and (not with_real_transient):
                only_neuropil = False
                with_cropped_real_transient = True
                self.cropped_transient_movies["real"].append(movie_data)
                n_cropped_transient = movie_info["n_cropped_transient"]
                self.n_cropped_transient_dict["real"][n_cropped_transient] = \
                    self.n_cropped_transient_dict["real"].get(n_cropped_transient, 0) + n_movies
                if "transients_amplitudes" in movie_info:
                    sorted_amplitudes["real"].append(np.max(movie_info["transients_amplitudes"]))
                    amplitudes_movie_index["real"].append(movie_index)
            if ("n_fake_transient" in movie_info) and (not with_real_transient) and (not with_cropped_real_transient):
                only_neuropil = False
                with_fake_transient = True
                self.transient_movies["fake"].append(movie_data)
                n_fake_transient = movie_info["n_fake_transient"]
                if n_fake_transient == 1:
                    self.full_1_transient["fake"].append(movie_data)
                else:
                    self.full_2p_transient["fake"].append(movie_data)
                self.n_transient_dict["fake"][n_fake_transient] = \
                    self.n_transient_dict["fake"].get(n_fake_transient, 0) + n_movies

                if "fake_transients_amplitudes" in movie_info:
                    self.transient_amplitudes["fake"].extend(movie_info["fake_transients_amplitudes"])
                    sorted_amplitudes["fake"].append(np.max(movie_info["fake_transients_amplitudes"]))
                    amplitudes_movie_index["fake"].append(movie_index)
                if "fake_transients_lengths" in movie_info:
                    self.transient_lengths["fake"].extend(movie_info["fake_transients_lengths"])
            if ("n_cropped_fake_transient" in movie_info) and (not with_real_transient) and (not with_fake_transient) \
                    and (not with_cropped_real_transient):
                only_neuropil = False
                self.cropped_transient_movies["fake"].append(movie_data)
                n_cropped_fake_transient = movie_info["n_cropped_fake_transient"]
                self.n_cropped_transient_dict["fake"][n_cropped_fake_transient] = \
                    self.n_cropped_transient_dict["fake"].get(n_cropped_fake_transient, 0) + n_movies
                if "fake_transients_amplitudes" in movie_info:
                    sorted_amplitudes["fake"].append(np.max(movie_info["fake_transients_amplitudes"]))
                    amplitudes_movie_index["fake"].append(movie_index)

            if with_fake_transient and with_real_transient:
                self.n_real_and_fake_transient += n_movies

            if only_neuropil:
                self.n_only_neuropil += n_movies
                self.neuropil_movies.append(movie_data)
            # TODO: See to replace this by a field category that could be encoded in cinac
            self.n_movies_by_session[movie_data.session_id] = \
                self.n_movies_by_session.get(movie_data.session_id, 0) + n_movies
            # self.n_movies_by_age[movie_data.ms.age] = self.n_movies_by_age.get(movie_data.ms.age, 0) + n_movies

        # sorting movie_data by amplitude, from the smallest to the biggest
        for which_ones in ["real", "fake"]:
            index_array = np.argsort(sorted_amplitudes[which_ones])
            for index in index_array:
                self.full_transient_sorted_amplitude[which_ones].append(amplitudes_movie_index[which_ones][index])

        self.only_neuropil_perc = (self.n_only_neuropil / self.n_movie_patch) * 100

        if self.debug_mode:
            print(f"{'#' * 10}")
            print(f"{'#' * 10}")
            print(f"{'#' * 10}")
            print(f"len train data {self.n_movie_patch}")
            print("")
            print(f"n_real_and_fake_transient {self.n_real_and_fake_transient}")
            print("")
            print("")

        for which_ones in ["real", "fake"]:
            # calculating the max number of augmentations done
            movies = self.transient_movies[which_ones]
            movies.extend(self.cropped_transient_movies[which_ones])
            transformations_done = []
            for movie in movies:
                transformations_done.append(movie.n_augmentations_to_perform)
            transformations_done = np.array(transformations_done)
            if self.debug_mode and (len(transformations_done) > 0):
                print(f"{which_ones}: transformations, min {np.min(transformations_done)}, "
                      f"max {np.max(transformations_done)}, "
                      f"mean {str(np.round(np.mean(transformations_done), 2))}")
            if self.debug_mode:
                print(f"{which_ones}: n_transient_dict {self.n_transient_dict[which_ones]}")
            self.n_full_transient_total[which_ones] = 0
            self.n_full_1_transient[which_ones] = 0
            self.n_full_2p_transient[which_ones] = 0
            for rep, count in self.n_transient_dict[which_ones].items():
                self.n_full_transient_total[which_ones] += count
                if rep == 1:
                    self.n_full_1_transient[which_ones] += count
                else:
                    self.n_full_2p_transient[which_ones] += count
            if self.n_full_transient_total[which_ones] > 0:
                self.full_1_transient_perc[which_ones] = (self.n_full_1_transient[which_ones] /
                                                          self.n_full_transient_total[which_ones]) * 100
                self.full_2p_transient_perc[which_ones] = (self.n_full_2p_transient[which_ones] /
                                                           self.n_full_transient_total[which_ones]) * 100
                if self.debug_mode:
                    print(perc_color + f"1 full {which_ones} transient perc: "
                    f"{str(np.round(self.full_1_transient_perc[which_ones], 2))} %" + reset_color)
                    print(perc_color + f"2+ full {which_ones} transient perc: "
                    f"{str(np.round(self.full_2p_transient_perc[which_ones], 2))} %" + reset_color)
            if self.debug_mode:
                print(f"{which_ones}: n_cropped_transient_dict {self.n_cropped_transient_dict[which_ones]}")
            self.n_cropped_transient_total[which_ones] = 0
            for rep, count in self.n_cropped_transient_dict[which_ones].items():
                self.n_cropped_transient_total[which_ones] += count

            self.n_transient_total[which_ones] = self.n_cropped_transient_total[which_ones] + \
                                                 self.n_full_transient_total[which_ones]
            self.total_transient_perc[which_ones] = (self.n_transient_total[which_ones] / self.n_movie_patch) * 100

            if self.n_transient_total[which_ones] > 0:
                self.full_transient_perc[which_ones] = (self.n_full_transient_total[which_ones] /
                                                        self.n_transient_total[which_ones]) * 100
                self.cropped_transient_perc[which_ones] = (self.n_cropped_transient_total[which_ones] /
                                                           self.n_transient_total[which_ones]) * 100
                if self.debug_mode:
                    print(perc_color + f"Full {which_ones}: "
                    f"{str(np.round(self.full_transient_perc[which_ones], 2))} %" + reset_color)
                    print(perc_color + f"Cropped {which_ones}: "
                    f"{str(np.round(self.cropped_transient_perc[which_ones], 2))} %" + reset_color)
            if self.debug_mode and (len(self.transient_lengths[which_ones]) > 0):
                print(f"{which_ones}: transient_lengths n {len(self.transient_lengths[which_ones])} / "
                      f"min-max {np.min(self.transient_lengths[which_ones])} - "
                      f"{np.max(self.transient_lengths[which_ones])}")
                print(f"{which_ones}: mean transient_amplitudes {np.mean(self.transient_amplitudes[which_ones])}")
                print("")
                print("")

        if self.debug_mode:
            for which_ones in ["real", "fake"]:
                print(perc_color + f"Total movie with {which_ones} transients {self.n_transient_total[which_ones]}: "
                f"{str(np.round(self.total_transient_perc[which_ones], 2))} %" + reset_color)
            print(perc_color + f"n_only_neuropil {self.n_only_neuropil}: "
            f"{str(np.round(self.only_neuropil_perc, 2))} %" + reset_color)
            print("")
            print("")

        # if self.debug_mode:
            print(f"n_movies_by_session {self.n_movies_by_session}")
            for session, count in self.n_movies_by_session.items():
                print(perc_color + f"{session}: {str(np.round((count / self.n_movie_patch) * 100, 2))} %" + reset_color)
        #     print(f"n_movies_by_age {self.n_movies_by_age}")
        #     for age, count in self.n_movies_by_age.items():
        #         print(perc_color + f"p{age}: {str(np.round((count / self.n_movie_patch) * 100, 2))} %" + reset_color)

    def add_augmentation_to_all_patches(self, n_augmentation):
        """
        Add to all movie patches in the camember a given number of augmentation, except neuropil
        :param n_augmentation:
        :return:
        """
        for movie_patch_data in self.data_list:
            if "only_neuropil" not in movie_patch_data.movie_info:
                movie_patch_data.add_n_augmentation(n_augmentation)

    def set_weights(self):
        # first we compute the thresholds
        # print(f"len(self.transient_amplitudes['real']) {len(self.transient_amplitudes['real'])}")
        real_amplitudes = np.unique(self.transient_amplitudes["real"])
        # print(f"real_amplitudes {len(real_amplitudes)}")
        fake_amplitudes = np.unique(self.transient_amplitudes["fake"])
        real_lengths = np.unique(self.transient_lengths["real"])
        fake_lengths = np.unique(self.transient_lengths["fake"])

        if len(real_amplitudes) > 0:
            real_amplitudes_threshold = np.percentile(real_amplitudes, 10)
        else:
            real_amplitudes_threshold = None
        if len(fake_amplitudes) > 0:
            fake_amplitudes_threshold = np.percentile(fake_amplitudes, 90)
        else:
            fake_amplitudes_threshold = None
        if len(real_lengths) > 0:
            real_lengths_threshold = np.percentile(real_lengths, 90)
        else:
            real_lengths_threshold = None

        if len(fake_lengths) > 0:
            fake_lengths_threshold = np.percentile(fake_lengths, 90)
        else:
            fake_lengths_threshold = None

        for movie_data in self.data_list:
            movie_info = movie_data.movie_info
            if "n_transient" in movie_info:
                if movie_info["n_transient"] > 1:
                    movie_data.weight += 5
                if (real_lengths_threshold is not None) and ("transients_lengths" in movie_info):
                    lengths = np.array(movie_info["transients_lengths"])
                    if len(np.where(lengths > real_lengths_threshold)[0]) > 0:
                        # print(f"lengths {lengths}, real_lengths_threshold {real_lengths_threshold}")
                        # means at least a transient length is superior to the 90th percentile
                        movie_data.weight += 3
                if (real_amplitudes_threshold is not None) and ("transients_amplitudes" in movie_info):
                    amplitudes = np.array(movie_info["transients_amplitudes"])
                    if len(np.where(amplitudes < real_amplitudes_threshold)[0]) > 0:
                        # print(f"amplitudes {amplitudes}, real_amplitudes_threshold {real_amplitudes_threshold}")
                        # means at least a transient amplitude is inferior to the 10th percentile
                        movie_data.weight += 3
                continue
            if "n_cropped_transient" in movie_info:
                continue
            if "n_fake_transient" in movie_info:
                movie_data.weight += 50
                if (fake_lengths_threshold is not None) and ("fake_transients_lengths" in movie_info):
                    lengths = np.array(movie_info["fake_transients_lengths"])
                    if len(np.where(lengths > fake_lengths_threshold)[0]) > 0:
                        # print(f"lengths {lengths}, real_lengths_threshold {fake_lengths_threshold}")
                        # means at least a transient length is superior to the 90th percentile
                        movie_data.weight += 5
                if (fake_amplitudes_threshold is not None) and ("fake_transients_amplitudes" in movie_info):
                    amplitudes = np.array(movie_info["fake_transients_amplitudes"])
                    if len(np.where(amplitudes < fake_amplitudes_threshold)[0]) > 0:
                        # means at least a transient amplitude is superior to the 90th percentile
                        movie_data.weight += 10
                continue
            if "n_cropped_fake_transient" in movie_info:
                movie_data.weight += 5

    def balance_all(self, main_ratio_balance, first_round):
        # if a ratio is put to one, then the class is untouched, but then the sum of other ratio should be equal to
        # 1 or set to -1 as well
        # main_ratio_balance = (0.6, 0.25, 0.15)
        if self.debug_mode:
            print("")
            print(f"$$$$$$$$$$$$$$$$$$$$$$ camembert.balance_all {main_ratio_balance} $$$$$$$$$$$$$$$$$$$$$$")
            print("")

        tolerance = 0.5

        # dealing with the case of the ratio is 0
        if (main_ratio_balance[0] == 0) and first_round:
            # then we delete all real transients
            new_data_list = []
            for movie_data in self.data_list:
                movie_info = movie_data.movie_info
                if "n_transient" in movie_info:
                    continue
                if "n_cropped_transient" in movie_info:
                    continue
                new_data_list.append(movie_data)
            self.data_list = new_data_list
            # updating the stat
            self.compute_slices()

        if (main_ratio_balance[1] == 0) and first_round:
            # then we delete all fake transients
            new_data_list = []
            for movie_data in self.data_list:
                movie_info = movie_data.movie_info
                if "n_transient" in movie_info:
                    new_data_list.append(movie_data)
                    continue
                if "n_cropped_transient" in movie_info:
                    new_data_list.append(movie_data)
                    continue
                if "n_fake_transient" in movie_info:
                    continue
                if "n_cropped_fake_transient" in movie_info:
                    continue
                new_data_list.append(movie_data)
            self.data_list = new_data_list
            # updating the stat
            self.compute_slices()

        if (main_ratio_balance[0] > 0) and (main_ratio_balance[1] > 0):

            ratio_real_fake = main_ratio_balance[0] / main_ratio_balance[1]

            if (self.n_transient_total["real"] > 0) and (self.n_transient_total["fake"] > 0):
                if np.abs((self.n_transient_total["real"] / self.n_transient_total["fake"]) - ratio_real_fake) > \
                        tolerance:
                    if (self.n_transient_total["fake"] * ratio_real_fake) > self.n_transient_total["real"]:
                        # it means they are too many fakes, we need to add real transients or delete fake ones
                        if first_round:
                            # we delete fake, as we don't care about balance in fakes during first_round
                            # we want to change self.data_list
                            n_fake_to_delete = int(self.n_transient_total["fake"] -
                                                   (self.n_transient_total["real"] / ratio_real_fake))
                            delete_low_amplitudes_first = True
                            new_data_list = []
                            if delete_low_amplitudes_first:
                                # but keeping some still, thus keeping one and removing one
                                indices_to_remove = []
                                n_fake_removed = 0
                                sorted_index = 0
                                while n_fake_removed < n_fake_to_delete:
                                    if sorted_index >= len(self.full_transient_sorted_amplitude["fake"]):
                                        break
                                    index_data_list = self.full_transient_sorted_amplitude["fake"][sorted_index]
                                    movie_data = self.data_list[index_data_list]
                                    if movie_data.to_keep_absolutely:
                                        # print("while n_fake_removed < n_fake_to_delete: movie_data.to_keep_absolutely")
                                        sorted_index += 1
                                        continue
                                    indices_to_remove.append(index_data_list)
                                    n_fake_removed += (1 + movie_data.n_augmentations_to_perform)

                                for index_data_list, movie_data in enumerate(self.data_list):
                                    if index_data_list in indices_to_remove:
                                        continue
                                    new_data_list.append(movie_data)
                            else:
                                n_fake_removed = 0
                                for movie_data in self.data_list:
                                    if movie_data.to_keep_absolutely:
                                        print("removing fake: movie_data.to_keep_absolutely")
                                        new_data_list.append(movie_data)
                                        continue
                                    movie_info = movie_data.movie_info
                                    if "n_transient" in movie_info:
                                        new_data_list.append(movie_data)
                                        continue
                                    if "n_cropped_transient" in movie_info:
                                        new_data_list.append(movie_data)
                                        continue
                                    if "n_fake_transient" in movie_info:
                                        if n_fake_removed < n_fake_to_delete:
                                            n_fake_removed += (1 + movie_data.n_augmentations_to_perform)
                                            continue
                                    if "n_cropped_fake_transient" in movie_info:
                                        if n_fake_removed < n_fake_to_delete:
                                            n_fake_removed += (1 + movie_data.n_augmentations_to_perform)
                                            continue
                                    new_data_list.append(movie_data)
                            self.data_list = new_data_list
                        else:
                            # we add real transients
                            n_real_to_add = (self.n_transient_total["fake"] * ratio_real_fake) - \
                                            self.n_transient_total["real"]
                            # print(f"n_real_to_add {n_real_to_add}")
                            # we want to add the same numbers such that we keep the ratio among the real_transients
                            # n_unique_real_transients represents the number of original movie patches, without taking in
                            # consideration the transformations that will be made
                            n_unique_real_transients = len(self.transient_movies["real"])
                            n_unique_real_transients += len(self.cropped_transient_movies["real"])
                            n_augmentations_options = [n_unique_real_transients * x for x in
                                                       np.arange(0, (self.n_max_transformations -
                                                                     self.min_augmentation_for_transient["real"]) + 3)]
                            n_augmentations_options = np.array(n_augmentations_options)
                            idx = (np.abs(n_augmentations_options - n_real_to_add)).argmin()
                            # print(f"idx {idx}, len(n_augmentations_options): {len(n_augmentations_options)}")
                            # print(f"n_augmentations_options[idx] {n_augmentations_options[idx]}")
                            if idx > 0:
                                n_transients_at_max_before = 0
                                n_transients_at_max_after = 0
                                n_added = 0
                                for movie_patch_data in self.transient_movies["real"]:
                                    augm_before = movie_patch_data.n_augmentations_to_perform
                                    movie_patch_data.add_n_augmentation(n_augmentation=idx)
                                    augm_after = movie_patch_data.n_augmentations_to_perform
                                    if augm_before == augm_after:
                                        n_transients_at_max_before += 1
                                    elif (augm_before + idx) < augm_after:
                                        n_transients_at_max_after += 1
                                    n_added += (augm_after - augm_before)
                                for movie_patch_data in self.cropped_transient_movies["real"]:
                                    augm_before = movie_patch_data.n_augmentations_to_perform
                                    movie_patch_data.add_n_augmentation(n_augmentation=idx)
                                    augm_after = movie_patch_data.n_augmentations_to_perform
                                    if augm_before == augm_after:
                                        n_transients_at_max_before += 1
                                    elif (augm_before + idx) < augm_after:
                                        n_transients_at_max_after += 1
                                    n_added += (augm_after - augm_before)
                                # TODO: if the ratio is not good, it means we couldn't add more transformation, in that
                                # TODO: case we want to remove some fake frames ?
                                if self.debug_mode:
                                    print(f"n_real_to_add {n_real_to_add}, n_added {n_added}")
                                    # print(f"n_transients_at_max_before {n_transients_at_max_before}, "
                                    #       f"n_transients_at_max_after {n_transients_at_max_before}")

                    else:
                        # it means they are too many real, we need to add fake transients
                        n_fake_to_add = (self.n_transient_total["real"] / ratio_real_fake) - \
                                        self.n_transient_total["fake"]
                        # we want to add the same numbers such that we keep the ratio among the real_transients
                        # n_unique_fake_transients represents the number of original movie patches, without taking in
                        # consideration the transformations that will be made
                        n_unique_fake_transients = len(self.transient_movies["fake"])
                        n_unique_fake_transients += len(self.cropped_transient_movies["fake"])
                        n_augmentations_options = [n_unique_fake_transients * x for x in
                                                   np.arange(1, (self.n_max_transformations -
                                                                 self.min_augmentation_for_transient["fake"]) + 2)]
                        n_augmentations_options = np.array(n_augmentations_options)
                        idx = (np.abs(n_augmentations_options - n_fake_to_add)).argmin()
                        if idx > 0:
                            for movie_patch_data in self.transient_movies["fake"]:
                                movie_patch_data.add_n_augmentation(n_augmentation=idx)
                            for movie_patch_data in self.cropped_transient_movies["fake"]:
                                movie_patch_data.add_n_augmentation(n_augmentation=idx)

            # updating the stat
            self.compute_slices()

        if (main_ratio_balance[2] == 0) and first_round:
            # then we delete all neuropil patches
            new_data_list = []
            for movie_data in self.data_list:
                movie_info = movie_data.movie_info
                if ("only_neuropil" in movie_info) and (not movie_data.to_keep_absolutely):
                    continue
                new_data_list.append(movie_data)
            self.data_list = new_data_list
        elif (main_ratio_balance[2] > 0) and (self.n_only_neuropil > 0):
            if (self.n_transient_total["real"] > 0) or (self.n_transient_total["fake"] > 0):
                n_transient_total = self.n_transient_total["real"] + self.n_transient_total["fake"]
                ratio_transients_neuropils = (main_ratio_balance[0] + main_ratio_balance[1]) / main_ratio_balance[2]
                if np.abs((n_transient_total / self.n_only_neuropil) -
                          ratio_transients_neuropils) > tolerance:
                    if (self.n_only_neuropil * ratio_transients_neuropils) > n_transient_total:
                        # it means they are too many neuropil, we need to remove some
                        # TODO: See to remove neuropils with the lowest variation
                        n_neuropils_to_remove = int(self.n_only_neuropil -
                                                    (n_transient_total / ratio_transients_neuropils))
                        # print(f"!!!!!!!!!!!!!!!!!! n_neuropils_to_remove {n_neuropils_to_remove}")
                        # we want to change self.data_list
                        n_neuropils_removed = 0
                        new_data_list = []
                        for movie_data in self.data_list:
                            movie_info = movie_data.movie_info
                            if ("only_neuropil" in movie_info) and (n_neuropils_removed < n_neuropils_to_remove) and \
                                    (not movie_data.to_keep_absolutely):
                                n_neuropils_removed += 1 + movie_data.n_augmentations_to_perform
                            else:
                                new_data_list.append(movie_data)
                        self.data_list = new_data_list

                    else:
                        # it means they are too many transients, we need to add neuropil
                        if self.debug_mode:
                            print(f"=== adding neuropil")
                        neuropil_to_add = (n_transient_total / ratio_transients_neuropils) - self.n_only_neuropil
                        if self.debug_mode:
                            print(f"=== neuropil_to_add {neuropil_to_add}")
                            print(f"=== len(self.neuropil_movies) {len(self.neuropil_movies)}")
                        augmentation_added = 0
                        movie_index = 0
                        while augmentation_added < neuropil_to_add:
                            self.neuropil_movies[movie_index].add_n_augmentation(n_augmentation=1)
                            movie_index = (movie_index + 1) % len(self.neuropil_movies)
                            augmentation_added += 1

        if self.debug_mode:
            print("")
            print(f"***************** After balancing real transients, fake ones and neuropil *****************")
            print("")
        # updating the stat
        self.compute_slices()

    def balance_transients(self, which_ones, crop_non_crop_ratio_balance, non_crop_ratio_balance):
        # if a ratio is put to one, then the class is untouched, but then the sum of other ratio should be equal to
        # 1 or set to -1 as well
        if which_ones not in ["fake", "real"]:
            raise Exception(f"which_ones not in {['fake', 'real']}")

        if self.debug_mode:
            print("")
            print(f"$$$$$$$$$$$$$$$$$$$$$$ camembert.balance_transients {which_ones} $$$$$$$$$$$$$$$$$$$$$$")
            print("")

        if self.n_transient_total[which_ones] == 0:
            return

        tolerance = 0.5
        if self.min_augmentation_for_transient[which_ones] > 0:
            for movie_data in self.transient_movies[which_ones]:
                movie_data.add_n_augmentation(n_augmentation=self.min_augmentation_for_transient[which_ones])
        if self.min_augmentation_for_cropped_transient[which_ones] > 0:
            for movie_data in self.cropped_transient_movies[which_ones]:
                movie_data.add_n_augmentation(n_augmentation=self.min_augmentation_for_cropped_transient[which_ones])

        if self.debug_mode:
            print("")
            print(f"$$ After adding min augmentation $$")
            print("")
        # updating stat
        self.compute_slices()

        if non_crop_ratio_balance[0] == 0:
            # we want to delete all the full 1 transient
            new_data_list = []
            for movie_data in self.data_list:
                movie_info = movie_data.movie_info
                if movie_data.to_keep_absolutely:
                    print("non_crop_ratio_balance[0] == 0: movie_data.to_keep_absolutely")
                    new_data_list.append(movie_data)
                    continue
                if (which_ones == "fake") and ("n_fake_transient" in movie_info) \
                        and ("n_transient" not in movie_info) and ("n_cropped_transient" not in movie_info):
                    if movie_info["n_fake_transient"] == 1:
                        continue
                if (which_ones == "real") and ("n_transient" in movie_info):
                    if movie_info["n_transient"] == 1:
                        continue
                new_data_list.append(movie_data)
            self.data_list = new_data_list

            # updating stat
            self.compute_slices()
        if non_crop_ratio_balance[1] == 0:
            # we want to delete all the  2p full transient
            new_data_list = []
            for movie_data in self.data_list:
                movie_info = movie_data.movie_info
                if movie_data.to_keep_absolutely:
                    print("non_crop_ratio_balance[1] == 0: movie_data.to_keep_absolutely")
                    new_data_list.append(movie_data)
                    continue
                if (which_ones == "fake") and ("n_fake_transient" in movie_info) \
                        and ("n_transient" not in movie_info) and ("n_cropped_transient" not in movie_info):
                    if movie_info["n_fake_transient"] > 1:
                        continue
                if (which_ones == "real") and ("n_transient" in movie_info):
                    if movie_info["n_transient"] > 1:
                        continue
                new_data_list.append(movie_data)
            self.data_list = new_data_list

            # updating stat
            self.compute_slices()

        if (non_crop_ratio_balance[0] > 0) and (non_crop_ratio_balance[1] > 0):
            ratio = non_crop_ratio_balance[0] / non_crop_ratio_balance[1]
            if ((self.n_full_2p_transient[which_ones] > 0) and (self.n_full_1_transient[which_ones] > 0)) and \
                    (np.abs(
                        (self.n_full_1_transient[which_ones] / self.n_full_2p_transient[
                            which_ones]) - ratio) > tolerance):
                # we have a 5% tolerance
                if (self.n_full_2p_transient[which_ones] * ratio) > self.n_full_1_transient[which_ones]:
                    # to balance real, we augment them, to balance fake: we remove some
                    # TODO: take in consideration the initial balance and if possible augment fake to balance the session
                    if which_ones == "real":
                        # it means we don't have enough full 1 transient and need to augment self.n_full_1_transient
                        # first we need to determine the difference
                        full_1_to_add = (self.n_full_2p_transient[which_ones] * ratio) - self.n_full_1_transient[
                            which_ones]
                        if self.debug_mode:
                            print(f"n_full_2p_transient[which_ones] {self.n_full_2p_transient[which_ones]}, "
                                  f" ratio {ratio}, n_full_1_transient[which_ones] "
                                  f"{self.n_full_1_transient[which_ones]}")
                            print(f"diff {full_1_to_add}")
                        augmentation_added = 0
                        movie_index = 0
                        while augmentation_added < full_1_to_add:
                            self.full_1_transient[which_ones][movie_index].add_n_augmentation(n_augmentation=1)
                            movie_index = (movie_index + 1) % len(self.full_1_transient[which_ones])
                            augmentation_added += 1
                    else:
                        # we have too much full 2p transient, we want to remove some (for fakes ones)
                        # first we need to determine the difference
                        n_full_2p_to_remove = self.n_full_2p_transient[which_ones] - \
                                              (self.n_full_1_transient[which_ones] / ratio)
                        # we want to change self.data_list
                        n_full_2p_removed = 0
                        new_data_list = []
                        for movie_data in self.data_list:
                            movie_info = movie_data.movie_info
                            if movie_data.to_keep_absolutely:
                                print(" too much full 2p transient: movie_data.to_keep_absolutely")
                                new_data_list.append(movie_data)
                                continue
                            if "n_transient" in movie_info:
                                new_data_list.append(movie_data)
                                continue
                            if "n_cropped_transient" in movie_info:
                                new_data_list.append(movie_data)
                                continue
                            if "n_fake_transient" in movie_info:
                                n_fake_transient = movie_info["n_fake_transient"]
                                if n_fake_transient == 1:
                                    new_data_list.append(movie_data)
                                    continue
                                elif n_full_2p_removed < n_full_2p_to_remove:
                                    n_full_2p_removed += (1 + movie_data.n_augmentations_to_perform)
                                    continue
                            new_data_list.append(movie_data)
                        self.data_list = new_data_list
                else:
                    if which_ones == "real":
                        # it means we have too many full_1_transient, need to augment self.n_full_2p_transient
                        # first we want to respect the non_crop_ratio_balance
                        full_2p_to_add = (self.n_full_1_transient[which_ones] / ratio) - self.n_full_2p_transient[
                            which_ones]
                        augmentation_added = 0
                        movie_index = 0
                        missed_augm = 0
                        while augmentation_added < full_2p_to_add:
                            movie_patch = self.full_2p_transient[which_ones][movie_index]
                            movie_patch.add_n_augmentation(n_augmentation=1)
                            movie_index = (movie_index + 1) % len(self.full_2p_transient[which_ones])
                            augmentation_added += 1
                    else:
                        # we have too much full 1 transient, we want to remove some (for fakes ones)
                        # first we need to determine the difference
                        n_full_1_to_remove = self.n_full_1_transient[which_ones] - \
                                             (self.n_full_2p_transient[which_ones] * ratio)
                        # we want to change self.data_list
                        n_full_1_removed = 0
                        new_data_list = []
                        for movie_data in self.data_list:
                            movie_info = movie_data.movie_info
                            if "n_transient" in movie_info:
                                new_data_list.append(movie_data)
                                continue
                            if "n_cropped_transient" in movie_info:
                                new_data_list.append(movie_data)
                                continue
                            if "n_fake_transient" in movie_info:
                                n_fake_transient = movie_info["n_fake_transient"]
                                if n_fake_transient > 1:
                                    new_data_list.append(movie_data)
                                    continue
                                elif n_full_1_removed < n_full_1_to_remove:
                                    n_full_1_removed += (1 + movie_data.n_augmentations_to_perform)
                                    continue
                            new_data_list.append(movie_data)
                        self.data_list = new_data_list

        if self.debug_mode:
            print("")
            print(f"$$ After balancing non cropped {which_ones} transients $$")
            print("")
        # updating stat
        self.compute_slices()

        if self.n_cropped_transient_total[which_ones] == 0:
            return

        if crop_non_crop_ratio_balance[0] == 0:
            # we want to delete all the non cropped transients
            new_data_list = []
            for movie_data in self.data_list:
                movie_info = movie_data.movie_info
                if movie_data.to_keep_absolutely:
                    print("crop_non_crop_ratio_balance[0] == 0: movie_data.to_keep_absolutely")
                    new_data_list.append(movie_data)
                    continue
                if (which_ones == "fake") and ("n_fake_transient" in movie_info) \
                        and ("n_transient" not in movie_info) and ("n_cropped_transient" not in movie_info):
                    continue
                if (which_ones == "real") and ("n_transient" in movie_info):
                    continue
                new_data_list.append(movie_data)
            self.data_list = new_data_list

            # updating stat
            self.compute_slices()

        if crop_non_crop_ratio_balance[1] == 0:
            # we want to delete all the  cropped transients
            new_data_list = []
            for movie_data in self.data_list:
                movie_info = movie_data.movie_info
                if movie_data.to_keep_absolutely:
                    print("crop_non_crop_ratio_balance[1] == 0 == 0: movie_data.to_keep_absolutely")
                    new_data_list.append(movie_data)
                    continue
                if (which_ones == "fake") and ("n_cropped_fake_transient" in movie_info) \
                        and ("n_transient" not in movie_info) and ("n_cropped_transient" not in movie_info):
                    continue
                if (which_ones == "real") and ("n_cropped_transient" in movie_info) and \
                        ("n_transient" not in movie_info):
                    continue
                new_data_list.append(movie_data)
            self.data_list = new_data_list

            # updating stat
            self.compute_slices()

        if (crop_non_crop_ratio_balance[0] > 0) and (crop_non_crop_ratio_balance[1] > 0):
            # now we want to balance cropped and non cropped (full)
            ratio = crop_non_crop_ratio_balance[0] / crop_non_crop_ratio_balance[1]
            if np.abs((self.n_full_transient_total[which_ones] / self.n_cropped_transient_total[which_ones]) - ratio) > \
                    tolerance:
                if (self.n_cropped_transient_total[which_ones] * ratio) > self.n_full_transient_total[which_ones]:
                    if which_ones == "real":
                        # it means we have to many cropped one and we need to augment n_full_transient_total
                        # first we need to determine the difference
                        full_to_add = (self.n_cropped_transient_total[which_ones] * ratio) - \
                                      self.n_full_transient_total[which_ones]
                        # we compute how many full movie we have (self.n_full_transient_total[which_ones] contains the number
                        # of movies included all the transformation that will be perform) we want the number of unique original
                        # movies before transformations
                        n_full = len(self.full_1_transient[which_ones]) + len(self.full_2p_transient[which_ones])
                        # we want to add the same numbers such that we keep the ratio among the full_fake_transients
                        n_movie_patch_options = [n_full * x for x in
                                                 np.arange(0, (self.n_max_transformations -
                                                               self.min_augmentation_for_transient[which_ones]) + 3)]
                        n_movie_patch_options = np.array(n_movie_patch_options)
                        idx = (np.abs(n_movie_patch_options - full_to_add)).argmin()
                        # print(f"//// {which_ones}: self.n_cropped_transient_total[which_ones] "
                        #       f"{self.n_cropped_transient_total[which_ones]}, "
                        #       f" ratio {ratio}, self.n_full_transient_total[which_ones] "
                        #       f"{self.n_full_transient_total[which_ones]}")
                        # print(f"////  full_to_add {full_to_add}")
                        # print(f"////  n_movie_patch_options {n_movie_patch_options}")
                        # print(f"////  idx {idx}")
                        # print(f"////  n_movie_patch_options[idx] {n_movie_patch_options[idx]}")
                        if idx > 0:
                            for movie_patch_data in self.full_1_transient[which_ones]:
                                movie_patch_data.add_n_augmentation(n_augmentation=idx)
                            for movie_patch_data in self.full_2p_transient[which_ones]:
                                movie_patch_data.add_n_augmentation(n_augmentation=idx)
                    else:
                        # it means we have to many cropped one and we need to remove some
                        # first we need to determine the difference
                        n_cropped_to_remove = self.n_cropped_transient_total[which_ones] - \
                                              (self.n_full_transient_total[which_ones] / ratio)
                        # we want to change self.data_list
                        n_cropped_removed = 0
                        new_data_list = []
                        for movie_data in self.data_list:
                            if movie_data.to_keep_absolutely:
                                print("to many cropped one: movie_data.to_keep_absolutely")
                                new_data_list.append(movie_data)
                                continue
                            movie_info = movie_data.movie_info
                            if "n_transient" in movie_info:
                                new_data_list.append(movie_data)
                                continue
                            if "n_cropped_transient" in movie_info:
                                new_data_list.append(movie_data)
                                continue
                            if "n_fake_transient" in movie_info:
                                new_data_list.append(movie_data)
                                continue
                            if "n_cropped_fake_transient" in movie_info:
                                if n_cropped_removed < n_cropped_to_remove:
                                    n_cropped_removed += (1 + movie_data.n_augmentations_to_perform)
                                    continue
                            new_data_list.append(movie_data)
                        self.data_list = new_data_list
                else:
                    # for full transient, we don't want to remove some even for fake, otherwise it would unbalance
                    # the full1 and full2
                    # if means we have too many full transient, we need to augment n_cropped_transient_total
                    # first we want to respect the non_crop_ratio_balance
                    n_cropped_to_add = (self.n_full_transient_total[which_ones] / ratio) - \
                                       self.n_cropped_transient_total[which_ones]
                    augmentation_added = 0
                    movie_index = 0
                    print(f"++++++ n_cropped_to_add {n_cropped_to_add}")
                    print(f"n_full_transient_total {which_ones} {self.n_full_transient_total[which_ones]}, "
                          f"n_cropped_transient_total {which_ones}  {self.n_cropped_transient_total[which_ones]}")
                    while augmentation_added < n_cropped_to_add:
                        self.cropped_transient_movies[which_ones][movie_index].add_n_augmentation(n_augmentation=1)
                        movie_index = (movie_index + 1) % len(self.cropped_transient_movies[which_ones])
                        augmentation_added += 1

        if self.debug_mode:
            print("")
            print(f"$$ After balancing cropped and non cropped {which_ones} transients $$")
            print("")
        self.compute_slices()


class StratificationDataProcessor:

    def __init__(self, data_list, description, n_max_transformations, main_ratio_balance=(0.6, 0.25, 0.15),
                 crop_non_crop_ratio_balance=(0.9, 0.1), non_crop_ratio_balance=(0.6, 0.4),
                 debug_mode=False):
        self.data_list = data_list
        self.n_transformations_for_session = n_max_transformations // 3
        self.n_max_transformations = n_max_transformations - self.n_transformations_for_session

        # for each session, we make a camembert of the movie_patches of this session
        # and balance the patches in the session
        # then we will balance the session among themselves by adding the number of augmentation
        # for all the patches of a given session, thus keeping the balance in the data
        self.movie_patches_data_by_session = dict()
        for movie_data in self.data_list:
            if movie_data.session_id not in self.movie_patches_data_by_session:
                self.movie_patches_data_by_session[movie_data.session_id] = []
            self.movie_patches_data_by_session[movie_data.session_id].append(movie_data)

        # just to have the stat
        StratificationCamembert(data_list=self.data_list,
                                description=description,
                                n_max_transformations=self.n_max_transformations,
                                debug_mode=True)

        self.camembert_by_session = dict()
        for session, session_movie_data in self.movie_patches_data_by_session.items():
            self.camembert_by_session[session] = StratificationCamembert(data_list=session_movie_data,
                                                                         description=session,
                                                                         n_max_transformations=self.n_max_transformations,
                                                                         debug_mode=debug_mode)
        # #### First we want to balance each session in itself
        # a first step, would be to first balance the fake transients
        # then see how many transformations are needed to be added to real transients to get the right proportion
        # then look at neuropil and from neuropil decide if we want to delete some or do data augmentation in some of

        # for camembert in self.camembert_by_session.values():
        #     camembert.balance_all(main_ratio_balance=main_ratio_balance)

        # balancing the real transients, fake ones and neuropils among themselves
        for camembert in self.camembert_by_session.values():
            camembert.balance_all(main_ratio_balance=main_ratio_balance, first_round=True)

        # them
        for camembert in self.camembert_by_session.values():
            camembert.balance_transients(which_ones="fake", crop_non_crop_ratio_balance=crop_non_crop_ratio_balance,
                                         non_crop_ratio_balance=non_crop_ratio_balance)

        # balancing the real transients
        for camembert in self.camembert_by_session.values():
            camembert.balance_transients(which_ones="real", crop_non_crop_ratio_balance=crop_non_crop_ratio_balance,
                                         non_crop_ratio_balance=non_crop_ratio_balance)

        # balancing the real transients, fake ones and neuropils among themselves
        for camembert in self.camembert_by_session.values():
            camembert.balance_all(main_ratio_balance=main_ratio_balance, first_round=False)

        # ####  then balance session between themselves
        # taking the sessions with the most movies and using it as exemples
        balancing_sessions = False
        if balancing_sessions:
            max_movie_patch = 0
            for camembert in self.camembert_by_session.values():
                max_movie_patch = max(max_movie_patch, camembert.n_movie_patch)

            for camembert in self.camembert_by_session.values():
                if camembert.n_movie_patch == max_movie_patch:
                    continue
                # we need to find the multiplicator between 1 and (self.n_transformations_for_session +1)
                # that would give the closest count from the max
                n_movie_patch = camembert.n_movie_patch
                # list of potential movie patches in this session depending on the augmentation factor
                # from 1 (no transformation added) to (self.n_transformations_for_session + 1)
                n_movie_patch_options = [n_movie_patch * x for x in
                                         np.arange(1, (self.n_transformations_for_session + 2))]
                n_movie_patch_options = np.array(n_movie_patch_options)
                idx = (np.abs(n_movie_patch_options - max_movie_patch)).argmin()
                if idx > 0:
                    camembert.add_augmentation_to_all_patches(n_augmentation=idx)

        # updating the data list, in case some movie patches would have been deleted
        new_data_list = []
        for camembert in self.camembert_by_session.values():
            new_data_list.extend(camembert.data_list)
        self.data_list = new_data_list

        # just to have the stat
        if debug_mode:
            print(f"////////// AFTER balancing sessions //////////////")
        balanced_camembert = StratificationCamembert(data_list=self.data_list,
                                                     description=description + "_balanced",
                                                     n_max_transformations=self.n_max_transformations,
                                                     debug_mode=True)
        # setting the weight based on amplitudes and lengths of the transients
        # also adding weight to fake transients, and multiple real ones
        balanced_camembert.set_weights()

    def get_new_data_list(self):
        return self.data_list


class StratificationCellTypeDataProcessor:
    """
    Class used to stratified data for cell type
    So far we make it simple, the postulate being that the data has already been stratify by the user (given the same
    amount of pramidal and interneuron cells for exemple). So for now we just used it to do data augmentation on each
    movie patch and create a new dataset thus
    """
    def __init__(self, data_list, description, n_max_transformations, debug_mode=False):
        self.data_list = data_list
        # self.n_transformations_for_session = n_max_transformations // 3
        self.n_max_transformations = n_max_transformations


        # for each session, we make a camembert of the movie_patches of this session
        # and balance the patches in the session
        # then we will balance the session among themselves by adding the number of augmentation
        # for all the patches of a given session, thus keeping the balance in the data
        self.movie_patches_data_by_session = dict()
        for movie_data in self.data_list:
            if movie_data.session_id not in self.movie_patches_data_by_session:
                self.movie_patches_data_by_session[movie_data.session_id] = []
            self.movie_patches_data_by_session[movie_data.session_id].append(movie_data)

        # just to have the stat
        StratificationCellTypeCamembert(data_list=self.data_list,
                                        description=description,
                                        n_max_transformations=self.n_max_transformations,
                                        debug_mode=True)

        self.camembert_by_session = dict()
        for session, session_movie_data in self.movie_patches_data_by_session.items():
            self.camembert_by_session[session] = StratificationCellTypeCamembert(data_list=session_movie_data,
                                                                                 description=session,
                                                                                 n_max_transformations=self.n_max_transformations,
                                                                                 debug_mode=debug_mode)

        # balancing the real transients, fake ones and neuropils among themselves
        for camembert in self.camembert_by_session.values():
            camembert.augment_them_all()

        # updating the data list, in case some movie patches would have been deleted
        new_data_list = []
        for camembert in self.camembert_by_session.values():
            new_data_list.extend(camembert.data_list)
        self.data_list = new_data_list

        # just to have the stat
        print(f"////////// AFTER balancing sessions //////////////")
        balanced_camembert = StratificationCellTypeCamembert(data_list=self.data_list,
                                                             description=description + "_balanced",
                                                             n_max_transformations=self.n_max_transformations,
                                                             debug_mode=True)

    def get_new_data_list(self):
        return self.data_list